from collections.abc import Awaitable, Callable
from typing import TypeVar, overload

from aidial_interceptors_sdk.utils.not_given import NOT_GIVEN, NotGiven

T = TypeVar("T")
P = TypeVar("P")


@overload
async def traverse_dict_value(
    path: P,
    d: dict,
    key: str,
    on_value: Callable[
        [P, T | NotGiven | None],
        Awaitable[T | NotGiven | None],
    ],
) -> dict: ...


@overload
async def traverse_dict_value(
    path: P,
    d: NotGiven,
    key: str,
    on_value: Callable[
        [P, T | NotGiven | None],
        Awaitable[T | NotGiven | None],
    ],
) -> NotGiven: ...


@overload
async def traverse_dict_value(
    path: P,
    d: None,
    key: str,
    on_value: Callable[
        [P, T | NotGiven | None],
        Awaitable[T | NotGiven | None],
    ],
) -> None: ...


async def traverse_dict_value(
    path: P,
    d: dict | NotGiven | None,
    key: str,
    on_value: Callable[
        [P, T | NotGiven | None],
        Awaitable[T | NotGiven | None],
    ],
) -> dict | NotGiven | None:
    if d is None or isinstance(d, NotGiven):
        return d

    old_value = d.get(key, NOT_GIVEN)
    new_value = await on_value(path, old_value)

    if new_value is NOT_GIVEN:
        if old_value is NOT_GIVEN:
            return d
        else:
            return {k: v for k, v in d.items() if k != key}
    else:
        return {**d, key: new_value}


@overload
async def traverse_required_dict_value(
    path: P,
    d: None,
    key: str,
    on_value: Callable[[P, T], Awaitable[T]],
) -> None: ...


@overload
async def traverse_required_dict_value(
    path: P,
    d: NotGiven,
    key: str,
    on_value: Callable[[P, T], Awaitable[T]],
) -> NotGiven: ...


@overload
async def traverse_required_dict_value(
    path: P,
    d: dict,
    key: str,
    on_value: Callable[[P, T], Awaitable[T]],
) -> dict: ...


async def traverse_required_dict_value(
    path: P,
    d: dict | NotGiven | None,
    key: str,
    on_value: Callable[[P, T], Awaitable[T]],
) -> dict | NotGiven | None:
    if d is None or isinstance(d, NotGiven):
        return d

    old_value = d.get(key)

    if old_value is None:
        raise ValueError(f"Missing required key {key!r} in a dictionary")

    new_value = await on_value(path, old_value)
    return {**d, key: new_value}


@overload
async def traverse_list(
    create_elem_path: Callable[[int], P],
    lst: NotGiven,
    on_elem: Callable[[P, T], Awaitable[list[T] | T]],
) -> NotGiven: ...


@overload
async def traverse_list(
    create_elem_path: Callable[[int], P],
    lst: None,
    on_elem: Callable[[P, T], Awaitable[list[T] | T]],
) -> None: ...


@overload
async def traverse_list(
    create_elem_path: Callable[[int], P],
    lst: list[T],
    on_elem: Callable[[P, T], Awaitable[list[T] | T]],
) -> list[T]: ...


async def traverse_list(
    create_elem_path: Callable[[int], P],
    lst: list[T] | NotGiven | None,
    on_elem: Callable[[P, T], Awaitable[list[T] | T]],
) -> list[T] | NotGiven | None:
    if lst is None or isinstance(lst, NotGiven):
        return lst

    ret: list[T] = []
    for idx, elem in enumerate(lst):
        idx = elem.get("index", idx) if isinstance(elem, dict) else idx
        elem = await on_elem(create_elem_path(idx), elem)
        if isinstance(elem, list):
            ret.extend(elem)
        else:
            ret.append(elem)

    return ret



from aidial_interceptors_sdk.examples.embeddings.blacklisted_words import (
    BlacklistedWordsInterceptor,
)
from aidial_interceptors_sdk.examples.embeddings.normalize_vector import (
    NormalizeVectorInterceptor,
)
from aidial_interceptors_sdk.examples.embeddings.project_vector import (
    ProjectVectorInterceptor,
)

__all__ = [
    "BlacklistedWordsInterceptor",
    "NormalizeVectorInterceptor",
    "ProjectVectorInterceptor",
]



from typing import Literal

from typing_extensions import override


class NotGiven:
    """
    A sentinel singleton class used to distinguish omitted keyword arguments
    from those passed in with the value None
    (which may have different behavior).

    Copied from https://github.com/openai/openai-python/blob/40f4cdb52a7494472c32e26c70f54bb41bb2bb57/src/openai/_types.py#L107-L133
    """

    def __bool__(self) -> Literal[False]:
        return False

    @override
    def __repr__(self) -> str:
        return "NOT_GIVEN"


NOT_GIVEN = NotGiven()



import math

import numpy as np
from PIL import Image, ImageDraw, ImageFont
from PIL.Image import Image as ImageObject


def _find_seamless_crop(image: ImageObject) -> tuple[int, int, int, int]:
    width, height = image.size

    image_gray = image.convert("L")
    pixels = np.array(image_gray)

    min_diff = np.inf
    min_x = -1
    x_diffs = []

    for x in range(10, width):
        diff = np.abs(pixels[:, 0] - pixels[:, x]).sum()
        x_diffs.append(int(diff))
        if diff < min_diff:
            min_diff = diff
            min_x = x

    vertical_seam = min_x

    min_diff = np.inf
    min_y = -1
    y_diffs = []

    for y in range(10, height):
        diff = np.abs(pixels[0, :] - pixels[y, :]).sum()
        y_diffs.append(int(diff))
        if diff < min_diff:
            min_diff = diff
            min_y = y

    horizontal_seam = min_y

    return (0, 0, vertical_seam, horizontal_seam)


def _create_watermark_texture(
    *,
    text: str,
    size: int = 800,
    font_size: int = 60,
    angle: int = 45,
) -> ImageObject:
    padded_size = int(size * math.sqrt(2))
    image_size = (padded_size, padded_size)
    background_color = (255, 255, 255, 0)
    image = Image.new("RGBA", image_size, background_color)

    draw = ImageDraw.Draw(image)

    font_size = 60
    font = ImageFont.load_default(font_size)

    text = "EPAM DIAL"
    text_color = (0, 0, 0, 128)  # Semi-transparent black

    text_width = int(draw.textlength(text, font=font))
    text_height = font_size

    text_layer = Image.new("RGBA", image_size, (255, 255, 255, 0))
    text_draw = ImageDraw.Draw(text_layer)

    for i in range(-50, 50):
        for j in range(-50, 50):
            text_position = (
                i * (text_width + 5),
                j * (text_height + 0),
            )

            text_draw.text(text_position, text, font=font, fill=text_color)

    text_layer = text_layer.rotate(angle, expand=0)

    image.paste(text_layer, (0, 0), text_layer)

    margin = (padded_size - size) // 2
    return image.crop(
        (margin, margin, padded_size - margin, padded_size - margin)
    )


def gen_watermark_image(text: str) -> ImageObject:
    image = _create_watermark_texture(text=text)
    crop_box = _find_seamless_crop(image)
    return image.crop(crop_box)


if __name__ == "__main__":
    gen_watermark_image("EPAM DIAL").save("watermark.png", "PNG")



import json
import logging
from collections.abc import AsyncIterator
from typing import cast

from aidial_sdk.chat_completion import ChatCompletion as DialChatCompletion
from aidial_sdk.chat_completion import Request as DialRequest
from aidial_sdk.chat_completion import Response as DialResponse
from aidial_sdk.chat_completion.chunks import DefaultChunk
from aidial_sdk.deployment.configuration import ConfigurationRequest
from aidial_sdk.exceptions import HTTPException as DialException
from aidial_sdk.exceptions import ResourceNotFoundError
from openai import AsyncStream
from openai.types.chat.chat_completion import ChatCompletion
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk

from aidial_interceptors_sdk.chat_completion.annotated_value import (
    AnnotatedException,
    Annotation,
)
from aidial_interceptors_sdk.chat_completion.base import (
    ChatCompletionInterceptor,
    RequestDict,
)
from aidial_interceptors_sdk.dial_client import DialClient
from aidial_interceptors_sdk.error import EarlyStreamExit
from aidial_interceptors_sdk.utils._debug import debug_logging
from aidial_interceptors_sdk.utils._dial_sdk import (
    cleanup_interceptor_configuration,
    parse_interceptor_configuration,
)
from aidial_interceptors_sdk.utils._exceptions import dial_exception_decorator
from aidial_interceptors_sdk.utils._http_client import HTTPClientFactory
from aidial_interceptors_sdk.utils._reflection import call_with_extra_body
from aidial_interceptors_sdk.utils.streaming import (
    block_response_to_streaming_chunk,
    map_stream,
    materialize_streaming_errors,
    singleton_stream,
)

_log = logging.getLogger(__name__)


def interceptor_to_chat_completion(
    cls: type[ChatCompletionInterceptor],
    dial_url: str,
    client_factory: HTTPClientFactory,
) -> DialChatCompletion:
    class Impl(DialChatCompletion):
        @dial_exception_decorator
        async def configuration(self, request: ConfigurationRequest) -> dict:
            if (schema := await cls.configuration_schema()) is None:
                raise ResourceNotFoundError(
                    "Configuration endpoint isn't implemented"
                )
            return schema.schema()

        @dial_exception_decorator
        async def chat_completion(
            self, request: DialRequest, response: DialResponse
        ) -> None:
            dial_client = await DialClient.create(
                dial_url=dial_url,
                api_key=request.api_key,
                api_version=request.api_version,
                authorization=request.jwt,
                headers=request.headers,
                client_factory=client_factory,
            )

            configuration_schema = await cls.configuration_schema()
            configuration = parse_interceptor_configuration(
                request, configuration_schema
            )

            interceptor = cls(
                dial_client=dial_client,
                request=request,
                response=response,
                configuration=configuration,
                **request.original_request.path_params,
            )

            request_body = await request.original_request.json()
            request_body = await debug_logging("request")(
                interceptor.traverse_request
            )(request_body)
            request_body = cleanup_interceptor_configuration(request_body)

            try:
                await interceptor.on_stream_start()

                def call_upstream(context: Annotation, request: dict):
                    return call_single_upstream(dial_client, context, request)

                async for value in await interceptor.call_upstreams(
                    request_body, call_upstream
                ):
                    if isinstance(value, AnnotatedException):
                        await interceptor.on_stream_error(value)
                    else:
                        await interceptor.traverse_response_chunk(value)

                await interceptor.on_stream_end()
            except EarlyStreamExit:
                pass

    return Impl()


async def call_single_upstream(
    dial_client: DialClient, context: Annotation, request: RequestDict
) -> AsyncIterator[dict | DialException]:
    response = cast(
        AsyncStream[ChatCompletionChunk] | ChatCompletion,
        await call_with_extra_body(
            dial_client.client.chat.completions.create, request
        ),
    )

    if isinstance(response, ChatCompletion):
        resp = response.to_dict()
        if _log.isEnabledFor(logging.DEBUG):
            _log.debug(f"upstream response[{context}]: {json.dumps(resp)}")

        # Non-streaming mode:
        # Removing the default fields which are generated by
        # DIAL SDK automatically.
        # It also means that these fields aren't proxied from the upstream.
        # They are recreated on each interceptor call.
        # If the fields aren't removed, then they will be merged
        # recursively with the one generated by SDK and we will end up with
        # "object": "chat.completionchat.completionchat.completion"
        for key in DefaultChunk.__annotations__:
            resp.pop(key, None)

        chunk = block_response_to_streaming_chunk(resp)
        stream = singleton_stream(chunk)
    else:
        # Streaming mode:
        # No need to remove default fields, because
        # they will be automatically overridden by the default fields
        # generated by DIAL SDK, when each chunk is merged naively with
        # a default chunk.

        def on_upstream_chunk(chunk: ChatCompletionChunk) -> dict:
            d = chunk.to_dict()
            if _log.isEnabledFor(logging.DEBUG):
                _log.debug(f"upstream chunk[{context}]: {json.dumps(d)}")
            return d

        stream = map_stream(on_upstream_chunk, response)

    return materialize_streaming_errors(stream)



from abc import ABC
from typing import Any

from aidial_sdk.exceptions import HTTPException as DialException
from aidial_sdk.pydantic_v1 import BaseModel

Annotation = Any | None


class AnnotatedValueBase(BaseModel, ABC):
    class Config:
        arbitrary_types_allowed = True

    annotation: Annotation = None


class AnnotatedChunk(AnnotatedValueBase):
    chunk: dict


class AnnotatedException(AnnotatedValueBase):
    error: DialException


AnnotatedValue = AnnotatedChunk | AnnotatedException



import logging

from google.cloud.dlp_v2 import DlpServiceAsyncClient, Likelihood

from aidial_interceptors_sdk.examples.utils.json import json_dumps_short

from .config import DeIdentificationConfig

_log = logging.getLogger(__name__)


class DlpClient:
    _parent: str
    _client: DlpServiceAsyncClient
    _config: DeIdentificationConfig

    def __init__(self, project: str, config: DeIdentificationConfig):
        self._parent = f"projects/{project}/locations/global"
        self._client = DlpServiceAsyncClient()
        self._config = config

    async def anonymize(self, input: str) -> str:
        info_types = [{"name": ty} for ty in self._config.info_types]

        transformation = {
            "info_types": info_types,
            "primitive_transformation": {"replace_with_info_type_config": {}},
        }

        deidentify_config = {
            "info_type_transformations": {"transformations": [transformation]},
            # Keeping potentially sensitive information if it was not possible to parse it.
            "transformation_error_handling": {"leave_untransformed": {}},
        }

        inspect_config = {
            "info_types": info_types,
            "min_likelihood": Likelihood.UNLIKELY,
        }

        item = {"value": input}

        request = {
            "parent": self._parent,
            "deidentify_config": deidentify_config,
            "inspect_config": inspect_config,
            "item": item,
        }

        if _log.isEnabledFor(logging.DEBUG):
            _log.debug(f"deidentify request: {json_dumps_short(request)}")

        response = await self._client.deidentify_content(request=request)

        if _log.isEnabledFor(logging.DEBUG):
            _log.debug(f"deidentify response: {json_dumps_short(response)}")

        return response.item.value



import contextlib
import functools
import logging

from aidial_sdk.exceptions import HTTPException as DialException
from openai import APIConnectionError, APIError, APIStatusError, APITimeoutError

_log = logging.getLogger(__name__)


def _parse_dial_exception(
    status_code: int,
    content: dict | str,
    headers: dict[str, str] | None = None,
):
    if (
        isinstance(content, dict)
        and (error := content.get("error"))
        and isinstance(error, dict)
    ):
        message = error.get("message") or "Unknown error"
        code = error.get("code")
        type = error.get("type")
        param = error.get("param")
        display_message = error.get("display_message")
    else:
        message = str(content)
        code = type = param = display_message = None

    return DialException(
        status_code=status_code,
        message=message,
        type=type,
        param=param,
        code=code,
        display_message=display_message,
        headers=headers,
    )


def to_dial_exception(exc: Exception) -> DialException:
    if isinstance(exc, APIStatusError):
        # Non-streaming errors reported by `openai` library via this exception

        r = exc.response
        headers = r.headers

        # The original content length may have changed
        # due to the response modification in the adapter.
        if "Content-Length" in headers:
            del headers["Content-Length"]

        # httpx library (used by openai) automatically sets
        # "Accept-Encoding:gzip,deflate" header in requests to the upstream.
        # Therefore, we may receive from the upstream gzip-encoded
        # response along with "Content-Encoding:gzip" header.
        # We either need to encode the response, or
        # remove the "Content-Encoding" header.
        if "Content-Encoding" in headers:
            del headers["Content-Encoding"]

        plain_headers = {k.decode(): v.decode() for k, v in headers.raw}

        try:
            content = r.json()
        except Exception:
            content = r.text

        return _parse_dial_exception(
            status_code=r.status_code,
            headers=plain_headers,
            content=content,
        )

    if isinstance(exc, APIError):
        # Streaming errors reported by `openai` library via this exception
        status_code: int = 500
        if exc.code:
            with contextlib.suppress(Exception):
                status_code = int(exc.code)

        return _parse_dial_exception(
            status_code=status_code,
            headers={},
            content={"error": exc.body or {}},
        )

    if isinstance(exc, APITimeoutError):
        return DialException("Request timed out", 504, "timeout")

    if isinstance(exc, APIConnectionError):
        return DialException(
            "Error communicating with OpenAI", 502, "connection"
        )

    if isinstance(exc, DialException):
        return exc

    return DialException(
        status_code=500,
        type="internal_server_error",
        message=str(exc),
    )


def dial_exception_decorator(func):
    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except Exception as e:
            dial_exception = to_dial_exception(e)
            _log.exception(
                f"Caught exception: {type(e).__module__}.{type(e).__name__}. "
                f"Converted to the DIAL exception: {dial_exception!r}"
            )
            raise dial_exception from e

    return wrapper



import logging
import os
import sys

from aidial_sdk import logger as aidial_logger
from uvicorn.logging import DefaultFormatter

LOG_LEVEL = os.getenv("LOG_LEVEL", "INFO")


def configure_loggers():
    # Making the uvicorn and dial sdk loggers delegate logging to the root logger
    for logger in [aidial_logger, logging.getLogger("uvicorn")]:
        logger.handlers = []
        logger.propagate = True

    # Setting up log levels
    for name in ["aidial_interceptors_sdk", "uvicorn"]:
        logging.getLogger(name).setLevel(LOG_LEVEL)

    # Configuring the root logger
    root = logging.getLogger()
    root.setLevel(LOG_LEVEL)

    root_has_stderr_handler = any(
        isinstance(handler, logging.StreamHandler)
        and handler.stream == sys.stderr
        for handler in root.handlers
    )

    # If stderr handler is already set, then no need to add another one
    if not root_has_stderr_handler:
        formatter = DefaultFormatter(
            fmt="%(levelprefix)s | %(asctime)s | %(name)s | %(process)d | %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
            use_colors=True,
        )

        handler = logging.StreamHandler(sys.stderr)
        handler.setFormatter(formatter)
        root.addHandler(handler)



from abc import ABC, abstractmethod

from .replacements import Replacements


class Anonymizer(ABC):
    @abstractmethod
    async def collect_replacements(
        self, text: str, *, replacements: Replacements | None = None
    ) -> Replacements:
        pass



from aidial_sdk.chat_completion import Request

from aidial_interceptors_sdk.chat_completion.element_path import ElementPath
from aidial_interceptors_sdk.chat_completion.helpers import (
    traverse_list,
    traverse_required_dict_value,
)
from aidial_interceptors_sdk.chat_completion.request_message_handler import (
    RequestMessageHandler,
)


class RequestHandler(RequestMessageHandler):
    request: Request

    @property
    def upstream_deployment_id(self) -> str | None:
        """
        Since ai-dial-core 0.43.0 returns ID of a DIAL deployment that
        this interceptor is assigned to.

        Not to be confused with the following values:

        1. self.request.deployment_id - the deployment ID of the interceptor itself.
            Populated from the `deployment_id` path variable in the endpoint that interceptor service is exposing:
            * POST /openai/deployments/{deployment_id}/chat/completions
            * POST /openai/deployments/{deployment_id}/embeddings

        2. self.request.model - the `model` field of the incoming Chat Completions request.
            There is no guarantee that it reflects an actual DIAL deployment ID.
        """
        return self.request.headers.get("X-DIAL-DEPLOYMENT-ID")

    async def on_request_message(
        self, path: ElementPath, message: dict
    ) -> list[dict]:
        return [message]

    async def on_request_messages(self, messages: list[dict]) -> list[dict]:
        return messages

    async def on_request(self, request: dict) -> dict:
        return request

    async def traverse_request(self, r: dict) -> dict:
        async def traverse_message(
            path: ElementPath, message: dict
        ) -> list[dict]:
            message = await self.traverse_request_message(path, message)
            return await self.on_request_message(path, message)

        async def traverse_messages(
            path: ElementPath, messages: list[dict]
        ) -> list[dict]:
            messages = await traverse_list(
                path.with_message_idx, messages, traverse_message
            )
            return await self.on_request_messages(messages)

        path = ElementPath()
        r = await traverse_required_dict_value(
            path, r, "messages", traverse_messages
        )
        r = await self.on_request(r)

        return r



from collections import defaultdict

from pydantic import BaseModel

from aidial_interceptors_sdk.examples.utils.markdown import MarkdownTable


class Replacements(BaseModel):
    replacements: dict[str, str] = {}
    """
    Map from an original value to its anonymized replacement in a text
    """

    indices: dict[str, int] = defaultdict(int)

    def get_replacement(self, entity: str, original_text: str) -> str:
        if original_text not in self.replacements:
            self.indices[entity] += 1
            index = self.indices[entity]
            self.replacements[original_text] = f"[{entity}-{index}]"

        return self.replacements[original_text]

    def is_empty(self) -> bool:
        return not bool(self.replacements)

    def to_markdown_table(self) -> str:
        table = MarkdownTable(
            title="Anonymized entities", headers=["Original", "Anonymized"]
        )
        for k, v in self.replacements.items():
            table.add_row([k, v])
        return table.to_markdown()

    def highlight_anonymized_entities(self, text: str) -> str:
        for v in self.replacements.values():
            text = text.replace(v, f"**{v}**")
        return text

    def deanonymize(self, text: str) -> str:
        for k, v in self.replacements.items():
            text = text.replace(v, k)
            # For cases when the model doesn't respect brackets
            text = text.replace(v[1:-1], k)
        return text

    def anonymize(self, text: str) -> str:
        for k, v in self.replacements.items():
            text = text.replace(k, v)
        return text



from aidial_interceptors_sdk.utils._env import get_env

from ..anonymizer.base import Anonymizer
from ..anonymizer.impl import AnonymizerInterceptor
from .anonymizer import GoogleDLPAnonymizer
from .config import GoogleDLPAnonymizerConfig


class GoogleDLPAnonymizerInterceptor(AnonymizerInterceptor):
    @classmethod
    async def configuration_schema(cls):
        return GoogleDLPAnonymizerConfig

    def get_anonymizer(self) -> Anonymizer:
        conf = self.get_configuration(GoogleDLPAnonymizerConfig)

        return GoogleDLPAnonymizer(
            get_env("GCP_PROJECT_ID"), conf.deidentification_config
        )



from abc import ABC, abstractmethod
from collections import defaultdict

from aidial_sdk.chat_completion import Stage
from typing_extensions import override

from aidial_interceptors_sdk.chat_completion.base import (
    ChatCompletionInterceptor,
)
from aidial_interceptors_sdk.chat_completion.element_path import ElementPath
from aidial_interceptors_sdk.examples.utils.markdown import MarkdownTable

from .base import Anonymizer
from .replacements import Replacements


class AnonymizerInterceptor(ChatCompletionInterceptor, ABC):
    # Request data
    request_n: int = 0
    anonymized_request: str = ""

    # Per-choice response data
    original_response_stages: dict[int, Stage] = {}
    content_buffers: dict[int, str] = defaultdict(str)

    replacements: Replacements = Replacements()

    @abstractmethod
    def get_anonymizer(self) -> Anonymizer:
        pass

    @override
    async def on_request_messages(self, messages: list[dict]) -> list[dict]:
        # Collect replacement dictionary first across all messages
        anonymizer = self.get_anonymizer()
        for message in messages:
            await anonymizer.collect_replacements(
                message.get("content") or "", replacements=self.replacements
            )

        # Then apply the replacements
        if not self.replacements.is_empty():
            chat_table = MarkdownTable(
                title="Anonymized chat history",
                headers=["Role", "Content"],
            )

            for message in messages:
                if message.get("content"):
                    message["content"] = self.replacements.anonymize(
                        message["content"]
                    )

                    content = self.replacements.highlight_anonymized_entities(
                        message["content"]
                    )

                    role = (message["role"] or "").upper()
                    chat_table.add_row([role, content])

            self.anonymized_request += (
                chat_table.to_markdown() + self.replacements.to_markdown_table()
            )

        return messages

    @override
    async def on_request(self, request: dict) -> dict:
        self.request_n = request.get("n") or 1
        return request

    @override
    async def on_stream_start(self) -> None:
        for choice_idx in range(self.request_n):
            with Stage(
                self.response._queue,
                choice_idx,
                self.reserve_stage_index(choice_idx),
                "Anonymized request",
            ) as stage:
                stage.append_content(self.anonymized_request)

            self.original_response_stages[choice_idx] = Stage(
                self.response._queue,
                choice_idx,
                self.reserve_stage_index(choice_idx),
                "Original response",
            )
            self.original_response_stages[choice_idx].open()

    @override
    async def on_response_choice(
        self, path: ElementPath, choice: dict
    ) -> list[dict] | dict:
        # NOTE: re-chunking invalidates streaming usage reported by the upstream model
        choice_idx = path.choice_idx
        assert choice_idx is not None

        if content := (choice.get("delta") or {}).get("content") or "":
            self.original_response_stages[choice_idx].append_content(content)

            buffer: str = self.content_buffers[choice_idx] + content

            br_open = buffer.count("[")
            br_closed = buffer.count("]")

            if br_open <= br_closed or choice.get("finish_reason"):
                choice["delta"]["content"] = self.replacements.deanonymize(
                    buffer
                )
                buffer = ""
            else:
                choice["delta"]["content"] = ""

            self.content_buffers[choice_idx] = buffer

        return choice

    async def on_stream_end(self) -> None:
        for choice_idx in range(self.request_n):
            if stage := self.original_response_stages.get(choice_idx):
                if content := self.content_buffers[choice_idx]:
                    stage.append_content(content)
                stage.close()



import io
import logging
from collections.abc import Mapping
from urllib.parse import urljoin

import httpx
from aidial_sdk.pydantic_v1 import BaseModel

_log = logging.getLogger(__name__)


class FileStorage(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    dial_url: str
    api_key: str
    http_client: httpx.AsyncClient

    @property
    def headers(self) -> Mapping[str, str]:
        return {"api-key": self.api_key}

    async def upload(
        self, url: str, content_type: str | None, content: bytes
    ) -> None:
        if self.to_dial_url(url) is None:
            raise ValueError(f"URL isn't DIAL url: {url!r}")
        url = self._to_abs_url(url)

        response = await self.http_client.put(
            url=url,
            files={"file": (url, io.BytesIO(content), content_type)},
            headers=self.headers,
        )
        response.raise_for_status()

        meta = response.json()

        _log.debug(f"uploaded file: url={url!r}, metadata={meta}")

    def to_dial_url(self, link: str) -> str | None:
        url = self._to_abs_url(link)
        base_url = f"{self.dial_url}/v1/"
        if url.startswith(base_url):
            return url.removeprefix(base_url)
        return None

    def _to_abs_url(self, link: str) -> str:
        base_url = f"{self.dial_url}/v1/"
        return urljoin(base_url, link)

    async def download(self, url: str) -> bytes:
        if self.to_dial_url(url) is None:
            raise ValueError(f"URL isn't DIAL url: {url!r}")
        url = self._to_abs_url(url)

        response = await self.http_client.get(url, headers=self.headers)
        response.raise_for_status()
        return response.content



from __future__ import annotations

import logging
import re

from aidial_sdk.pydantic_v1 import BaseModel

from .replacements import Replacements

_log = logging.getLogger(__name__)


class Replacement(BaseModel):
    entity_type: str
    idx: int

    def print(self):
        return f"[{self.entity_type}-{self.idx}]"

    @classmethod
    def parse(cls, text: str) -> Replacement | None:
        regexp = r"\[(\w+)-(\d+)\]"
        match = re.match(regexp, text)
        if match:
            return cls(entity_type=match.group(1), idx=int(match.group(2)))
        return None


def parse_anonymized_string(entities: list[str], s: str) -> list[str | int]:
    ret = []

    def _append(x: str | int):
        if isinstance(x, int) or (isinstance(x, str) and x):
            ret.append(x)

    i = 0
    n = len(s)
    last_match = 0

    while i < n:
        found = False
        if s[i] == "[":
            for j, e in enumerate(entities):
                if s[i + 1 : i + len(e) + 2] == e + "]":
                    _append(s[last_match:i])
                    _append(j)
                    i += len(e) + 2
                    last_match = i
                    found = True
                    break

        if not found:
            i += 1

    _append(s[last_match:])
    return ret


def create_indexed_replacements(
    replacements: Replacements,
    entities: list[str],
    original: str,
    anonymized: str,
) -> Replacements | None:
    anonymized_parsed = parse_anonymized_string(entities, anonymized)

    entity_indices = []
    regexp = ""

    for segment in anonymized_parsed:
        if isinstance(segment, int):
            regexp += "(.*?)"
            entity_indices.append(segment)
        else:
            regexp += re.escape(segment)

    match = re.fullmatch(regexp, original)
    if not match:
        _log.debug("Replacement parsing was unsuccessful")
        return None

    groups = match.groups()
    if len(groups) != len(entity_indices):
        _log.debug(
            f"Number of group matches ({len(groups)}) doesn't "
            f"match the expected number ({len(entity_indices)})"
        )
        return None

    for original_text, entity_index in zip(
        groups, entity_indices, strict=False
    ):
        replacements.get_replacement(entities[entity_index], original_text)

    return replacements



from datetime import datetime

from aidial_sdk.pydantic_v1 import BaseModel
from langfuse import Langfuse


class LangfuseClient(BaseModel):
    session_id: str
    tags: list[str]
    request_messages: list
    response_message: dict
    model_name: str = ""
    deployment_id: str = ""
    start_time: datetime
    end_time: datetime
    user_id: str | None
    metadata: dict = {}
    is_model: bool

    def connect(self):
        return Langfuse()

    def transmit(self):
        conn = self.connect()
        trace = conn.trace(
            name=self.model_name,
            session_id=self.session_id,
            tags=self.tags,
            input=self.request_messages,
            output=self.response_message,
            user_id=self.user_id,
            metadata=self.metadata,
        )
        if self.is_model:
            trace.generation(
                model=self.model_name,
                input=self.request_messages,
                output=self.response_message,
                start_time=self.start_time,
                end_time=self.end_time,
            )
        else:
            trace.span(
                name="Processing",
                input=self.request_messages,
                output=self.response_message,
                start_time=self.start_time,
                end_time=self.end_time,
            )
        conn.flush()
        return



"""
Utilities for pretty-printing JSON in debug logs.
These functions are useful for dumping large data structures,
with options to trim long strings and lists to specified limits.
"""

import json
from dataclasses import asdict, is_dataclass
from enum import Enum
from typing import Any

import proto
from pydantic.v1 import BaseModel as BaseModelV1

from aidial_interceptors_sdk.examples.utils.decorator import fail_safe
from aidial_interceptors_sdk.examples.utils.protobuf import message_to_dict


@fail_safe
def json_dumps_short(
    obj: Any, *, string_limit: int = 100, list_len_limit: int = 10, **kwargs
) -> str:
    def default(obj) -> str:
        return _truncate_strings(str(obj), string_limit)

    return json.dumps(
        _truncate_lists(
            _truncate_strings(_to_dict(obj, **kwargs), string_limit),
            list_len_limit,
        ),
        default=default,
    )


@fail_safe
def json_dumps(obj: Any, **kwargs) -> str:
    return json.dumps(_to_dict(obj, **kwargs), default=str)


def _to_dict(obj: Any, **kwargs) -> Any:
    def rec(val):
        return _to_dict(val, **kwargs)

    def dict_field(key: str, val: Any) -> Any:
        if key in kwargs.get("excluded_keys", []):
            return "<excluded>"
        return val

    if isinstance(obj, bytes):
        return f"<bytes>({len(obj):_} B)"

    if isinstance(obj, Enum):
        return obj.value

    if isinstance(obj, dict):
        filters = []
        if kwargs.get("exclude_none"):
            filters.append(lambda x: x is not None)
        if kwargs.get("exclude_empty_dict"):
            filters.append(lambda x: x != {})

        ret = {}
        for k, v in obj.items():
            v = rec(dict_field(k, v))
            if all(f(v) for f in filters):
                ret[k] = v
        return ret

    if isinstance(obj, list):
        return [rec(element) for element in obj]

    if isinstance(obj, tuple):
        return tuple(rec(element) for element in obj)

    if isinstance(obj, BaseModelV1):
        return rec(obj.dict())

    if isinstance(obj, proto.Message):
        return rec(message_to_dict(obj))

    if hasattr(obj, "to_dict"):
        return rec(obj.to_dict())

    if is_dataclass(type(obj)):
        return rec(asdict(obj))

    return obj


def _truncate_strings(obj: Any, limit: int) -> Any:
    def rec(val):
        return _truncate_strings(val, limit)

    if isinstance(obj, dict):
        return {key: rec(value) for key, value in obj.items()}

    if isinstance(obj, list):
        return [rec(element) for element in obj]

    if isinstance(obj, tuple):
        return tuple(rec(element) for element in obj)

    if isinstance(obj, str) and len(obj) > limit:
        skip = len(obj) - limit
        return (
            obj[: limit // 2] + f"...({skip:_} skipped)..." + obj[-limit // 2 :]
        )

    return obj


def _truncate_lists(obj: Any, limit: int) -> Any:
    def rec(val):
        return _truncate_lists(val, limit)

    if isinstance(obj, dict):
        return {key: rec(value) for key, value in obj.items()}

    if isinstance(obj, list):
        if len(obj) > limit:
            skip = len(obj) - limit
            obj = (
                obj[: limit // 2]
                + [f"...({skip:_} skipped)..."]
                + obj[-limit // 2 :]
            )
        return [rec(element) for element in obj]

    if isinstance(obj, tuple):
        return tuple(rec(element) for element in obj)

    return obj



from collections import OrderedDict
from collections.abc import Hashable
from typing import Generic, TypeVar

_K = TypeVar("_K", bound=Hashable)
_V = TypeVar("_V")


class LRUCache(Generic[_K, _V]):
    def __init__(self, maxsize: int = 128) -> None:
        self.cache: OrderedDict[_K, _V] = OrderedDict()
        self.maxsize: int = maxsize

    def lookup(self, key: _K) -> _V | None:
        if key in self.cache:
            self.cache.move_to_end(key)
            return self.cache[key]
        else:
            return None

    def save(self, key: _K, value: _V) -> None:
        if key in self.cache:
            self.cache.move_to_end(key)
        self.cache[key] = value
        if len(self.cache) > self.maxsize:
            self.cache.popitem(last=False)



from typing import Any, Self

from aidial_sdk.pydantic_v1 import BaseModel

from aidial_interceptors_sdk.chat_completion.index_mapper import IndexMapper


class ChoiceContext(BaseModel):
    index: int
    stage_index_mapper: IndexMapper[int]


class ElementPath(BaseModel):
    # Only for responses
    response_ctx: Any | None = None
    choice_ctx: ChoiceContext | None = None

    # Only for requests
    message_idx: int | None = None

    # Both for requests and responses
    stage_idx: int | None = None
    attachment_idx: int | None = None

    def with_response_ctx(self, response_ctx: Any) -> Self:
        return self.copy(update={"response_ctx": response_ctx})

    def with_choice_ctx(self, choice_ctx: ChoiceContext) -> Self:
        return self.copy(update={"choice_ctx": choice_ctx})

    def with_message_idx(self, message_idx: int) -> Self:
        return self.copy(update={"message_idx": message_idx})

    def with_stage_idx(self, stage_idx: int) -> Self:
        return self.copy(update={"stage_idx": stage_idx})

    def with_attachment_idx(self, attachment_idx: int) -> Self:
        return self.copy(update={"attachment_idx": attachment_idx})

    @property
    def choice_idx(self) -> int | None:
        return self.choice_ctx.index if self.choice_ctx else None

    @property
    def choice_stage_index_mapper(self) -> IndexMapper[int] | None:
        return self.choice_ctx.stage_index_mapper if self.choice_ctx else None



import base64

from aidial_sdk.exceptions import InvalidRequestError
from typing_extensions import override

from aidial_interceptors_sdk.chat_completion.base import (
    ChatCompletionInterceptor,
)
from aidial_interceptors_sdk.examples.utils.watermark.stamp import (
    stamp_watermark,
)


class ImageWatermarkInterceptor(ChatCompletionInterceptor):
    @override
    async def on_response_attachment(self, path, attachment: dict) -> dict:
        ty = attachment.get("type")

        if ty == "image/jpeg":
            format = "JPEG"
        elif ty == "image/png":
            format = "PNG"
        else:
            return attachment

        url = attachment.get("url")
        if url is not None:
            dial_url = self.dial_client.storage.to_dial_url(url)
            if dial_url is None:
                return attachment

            data = await self.dial_client.storage.download(url)
            data = stamp_watermark(data, format)

            # overwrite the original image
            await self.dial_client.storage.upload(url, ty, data)

        data = attachment.get("data")
        if data is not None:
            try:
                bytes = base64.b64decode(data)
            except Exception:
                raise InvalidRequestError(
                    "Attachment data isn't base64 encoded",
                ) from None

            bytes = stamp_watermark(bytes, format)
            attachment = {
                **attachment,
                "data": base64.b64encode(bytes).decode("utf-8"),
            }

        return attachment



from collections.abc import Hashable
from typing import Generic, TypeVar

from aidial_sdk.pydantic_v1 import BaseModel

_Index = TypeVar("_Index", bound=Hashable)


class IndexMapper(BaseModel, Generic[_Index]):
    """
    Used to maintain consistent mapping between indexed values in the incoming and outgoing streams, given that outgoing stream may include additional elements at fixed indices.
    """

    migrated: dict[_Index, int] = {}
    used_indices: set[int] = set()

    fresh_index: int = 0

    def reserve(self, index: int | None = None) -> int:
        if index is None:
            return self._get_fresh_index()

        if index in self.used_indices:
            raise ValueError(f"Index {index} is already taken")

        self.used_indices.add(index)
        return index

    def __call__(self, index: _Index) -> int:
        if index not in self.migrated:
            self.migrated[index] = self._get_fresh_index()
        return self.migrated[index]

    def _get_fresh_index(self) -> int:
        while self.fresh_index in self.used_indices:
            self.fresh_index += 1
        self.used_indices.add(self.fresh_index)
        return self.fresh_index



from typing_extensions import override

from aidial_interceptors_sdk.embeddings.base import EmbeddingsInterceptor
from aidial_interceptors_sdk.examples.utils.embedding_encoding import (
    base64_to_vector,
    vector_to_base64,
)


def project(vec: list[float], dim: int) -> list[float]:
    diff = dim - len(vec)
    if diff == 0:
        return vec
    elif diff > 0:
        return vec + [0.0] * diff
    else:
        return vec[:dim]


class ProjectVectorInterceptor(EmbeddingsInterceptor):
    dim: int

    @override
    async def modify_embedding(
        self, embedding: str | list[float]
    ) -> str | list[float]:
        if isinstance(embedding, str):
            vec = base64_to_vector(embedding)
            vec = project(vec, self.dim)
            return vector_to_base64(vec)
        else:
            return project(embedding, self.dim)



from aidial_sdk.pydantic_v1 import BaseModel

from aidial_interceptors_sdk.chat_completion.base import (
    ChatCompletionInterceptor,
    ChatCompletionNoOpInterceptor,
)
from aidial_interceptors_sdk.embeddings.base import (
    EmbeddingsInterceptor,
    EmbeddingsNoOpInterceptor,
)
from aidial_interceptors_sdk.examples.chat_completion import (
    BlacklistedWordsInterceptor as ChatBlacklistedWordsInterceptor,
)
from aidial_interceptors_sdk.examples.chat_completion import (
    CachingInterceptor as ChatCachingInterceptor,
)
from aidial_interceptors_sdk.examples.chat_completion import (
    GoogleDLPAnonymizerInterceptor,
    ImageWatermarkInterceptor,
    LangfuseInterceptor,
    PirateInterceptor,
    RejectExternalLinksInterceptor,
    ReplicatorInterceptor,
    SpacyAnonymizerInterceptor,
    StatisticsReporterInterceptor,
    WhitespaceAccumulatorInterceptor,
)
from aidial_interceptors_sdk.examples.embeddings import (
    BlacklistedWordsInterceptor as EmbeddingsBlacklistedWordsInterceptor,
)
from aidial_interceptors_sdk.examples.embeddings import (
    NormalizeVectorInterceptor,
    ProjectVectorInterceptor,
)


class Interceptors(BaseModel):
    chat_completions: dict[str, type[ChatCompletionInterceptor]] = {}
    embeddings: dict[str, type[EmbeddingsInterceptor]] = {}


EXAMPLE_INTERCEPTORS: Interceptors = Interceptors(
    chat_completions={
        "reply-as-pirate": PirateInterceptor,
        "reject-external-links": RejectExternalLinksInterceptor,
        "image-watermark": ImageWatermarkInterceptor,
        "statistics-reporter": StatisticsReporterInterceptor,
        "pii-anonymizer": SpacyAnonymizerInterceptor,
        "spacy-anonymizer": SpacyAnonymizerInterceptor,
        "google-dlp-anonymizer": GoogleDLPAnonymizerInterceptor,
        "replicator:{n:int}": ReplicatorInterceptor,
        "reject-blacklisted-words": ChatBlacklistedWordsInterceptor,
        "cache": ChatCachingInterceptor,
        "langfuse": LangfuseInterceptor,
        "no-op": ChatCompletionNoOpInterceptor,
        "whitespace-accumulator": WhitespaceAccumulatorInterceptor,
    },
    embeddings={
        "reject-blacklisted-words": EmbeddingsBlacklistedWordsInterceptor,
        "normalize-vector": NormalizeVectorInterceptor,
        "project-vector:{dim:int}": ProjectVectorInterceptor,
        "no-op": EmbeddingsNoOpInterceptor,
    },
)



import proto
from google.protobuf import json_format
from proto.marshal.collections import maps, repeated


def message_to_string(message: proto.Message) -> str:
    return json_format.MessageToJson(message._pb)


def message_to_dict(message: proto.Message) -> dict:
    return json_format.MessageToDict(message._pb)


# Borrowed from helper functions
# https://github.com/GoogleCloudPlatform/generative-ai/blob/main/gemini/function-calling/function_calling_data_structures.ipynb
def recurse_proto_repeated_composite(repeated_object):
    repeated_list = []
    for item in repeated_object:
        if isinstance(item, repeated.RepeatedComposite):
            item = recurse_proto_repeated_composite(item)
            repeated_list.append(item)
        elif isinstance(item, maps.MapComposite):
            item = recurse_proto_marshal_to_dict(item)
            repeated_list.append(item)
        else:
            repeated_list.append(item)

    return repeated_list


def recurse_proto_marshal_to_dict(marshal_object):
    new_dict = {}
    for k, v in marshal_object.items():
        if not v:
            continue
        elif isinstance(v, maps.MapComposite):
            v = recurse_proto_marshal_to_dict(v)
        elif isinstance(v, repeated.RepeatedComposite):
            v = recurse_proto_repeated_composite(v)
        new_dict[k] = v

    return new_dict



from .adapter import interceptor_to_chat_completion
from .base import ChatCompletionInterceptor, ChatCompletionNoOpInterceptor
from .element_path import ElementPath

__all__ = [
    "ChatCompletionInterceptor",
    "ChatCompletionNoOpInterceptor",
    "ElementPath",
    "interceptor_to_chat_completion",
]



import logging
from collections.abc import Callable

_log = logging.getLogger(__name__)


def fail_safe(func: Callable[..., str]) -> Callable[..., str]:
    def _wrapper(*args, **kwargs) -> str:
        try:
            return func(*args, **kwargs)
        except Exception as e:
            msg = f"Error executing {func.__name__}: {str(e)}"
            _log.warning(msg)
            return msg

    return _wrapper



from typing_extensions import override

from aidial_interceptors_sdk.embeddings.base import EmbeddingsInterceptor
from aidial_interceptors_sdk.examples.utils.embedding_encoding import (
    base64_to_vector,
    vector_to_base64,
)


def normalize(vec: list[float]) -> list[float]:
    norm = sum(x**2 for x in vec) ** 0.5
    return [x / norm for x in vec]


class NormalizeVectorInterceptor(EmbeddingsInterceptor):
    @override
    async def modify_embedding(
        self, embedding: str | list[float]
    ) -> str | list[float]:
        if isinstance(embedding, str):
            vec = base64_to_vector(embedding)
            vec = normalize(vec)
            return vector_to_base64(vec)
        else:
            return normalize(embedding)



import logging
import time

from typing_extensions import override

from aidial_interceptors_sdk.chat_completion.base import (
    ChatCompletionInterceptor,
)

_log = logging.getLogger(__name__)

_FLUSH_TIMEOUT = 1.0


class WhitespaceAccumulatorInterceptor(ChatCompletionInterceptor):
    last_send: float = 0.0
    cur_chunk: dict | None = None
    cur_chunks: int = 0
    timeout_sec: float = _FLUSH_TIMEOUT
    buffer: list[str] = []

    def flush(self) -> None:
        if self.cur_chunk is None:
            return
        if self.cur_chunks > 1:
            _log.info(f"flushed {self.cur_chunks} chunks")
        self.last_send = time.perf_counter()
        content = "".join(self.buffer)
        self.set_content(self.cur_chunk, content)
        self.send_chunk(self.cur_chunk)
        self.buffer.clear()
        self.cur_chunk = None
        self.cur_chunks = 0

    def need_flush(self, content: str) -> bool:
        passed = time.perf_counter() - self.last_send
        return content and not content.isspace() or (passed > self.timeout_sec)

    @staticmethod
    def set_content(chunk: dict, content: str):
        try:
            chunk["choices"][0]["delta"]["content"] = content
        except (KeyError, IndexError, TypeError):
            return None

    @staticmethod
    def parse_content_only_chunk(chunk: dict) -> str | None:
        if (choices := chunk.get("choices")) is None:
            return None

        if not isinstance(choices, list) or len(choices) != 1:
            return None

        choice = choices[0]
        if not isinstance(choice, dict) or set(choice.keys()) != {
            "finish_reason",
            "index",
            "delta",
        }:
            return None

        if choice["finish_reason"]:
            return None

        index = choice["index"]
        if not isinstance(index, int) or index != 0:
            return None

        delta = choice["delta"]
        if not isinstance(delta, dict) or set(delta.keys()) != {
            "content",
        }:
            return None

        content = delta["content"]
        if not isinstance(content, str):
            return None

        return content

    @override
    async def on_stream_start(self) -> None:
        self.last_send = time.perf_counter()

    @override
    async def on_stream_chunk(self, chunk: dict) -> None:
        if content := self.parse_content_only_chunk(chunk):
            self.cur_chunk = self.cur_chunk or chunk
            self.buffer.append(content)
            if self.need_flush(content):
                self.flush()
            else:
                self.cur_chunks += 1
        else:
            self.flush()
            self.send_chunk(chunk)

    @override
    async def on_stream_end(self) -> None:
        self.flush()



from aidial_interceptors_sdk.examples.chat_completion.blacklisted_words import (
    BlacklistedWordsInterceptor,
)
from aidial_interceptors_sdk.examples.chat_completion.cache import (
    CachingInterceptor,
)
from aidial_interceptors_sdk.examples.chat_completion.google_dlp_anonymiser import (
    GoogleDLPAnonymizerInterceptor,
)
from aidial_interceptors_sdk.examples.chat_completion.image_watermark import (
    ImageWatermarkInterceptor,
)
from aidial_interceptors_sdk.examples.chat_completion.langfuse import (
    LangfuseInterceptor,
)
from aidial_interceptors_sdk.examples.chat_completion.pii_anonymiser import (
    SpacyAnonymizerInterceptor,
)
from aidial_interceptors_sdk.examples.chat_completion.pirate import (
    PirateInterceptor,
)
from aidial_interceptors_sdk.examples.chat_completion.reject_external_links import (
    RejectExternalLinksInterceptor,
)
from aidial_interceptors_sdk.examples.chat_completion.replicator import (
    ReplicatorInterceptor,
)
from aidial_interceptors_sdk.examples.chat_completion.statistics_reporter import (
    StatisticsReporterInterceptor,
)
from aidial_interceptors_sdk.examples.chat_completion.whitespace_accumulator import (
    WhitespaceAccumulatorInterceptor,
)

__all__ = [
    "BlacklistedWordsInterceptor",
    "CachingInterceptor",
    "GoogleDLPAnonymizerInterceptor",
    "ImageWatermarkInterceptor",
    "LangfuseInterceptor",
    "SpacyAnonymizerInterceptor",
    "PirateInterceptor",
    "RejectExternalLinksInterceptor",
    "ReplicatorInterceptor",
    "StatisticsReporterInterceptor",
    "WhitespaceAccumulatorInterceptor",
]



import io
from functools import cache
from typing import Literal

from PIL import Image
from PIL.Image import Image as ImageObject

from aidial_interceptors_sdk.examples.utils.watermark.generate import (
    gen_watermark_image,
)


def _resize_tiling(image: ImageObject, box: tuple[int, int]) -> ImageObject:
    width, height = box
    tiled_image = Image.new("RGBA", (width, height), (0, 0, 0, 0))

    img_width, img_height = image.size

    for x in range(0, width + img_width, img_width):
        for y in range(0, height + img_height, img_height):
            tiled_image.paste(image, (x, y))

    return tiled_image


def _stamp_watermark_image(
    image: ImageObject, watermark: ImageObject
) -> ImageObject:
    tiled_watermark = _resize_tiling(watermark, image.size)

    if image.mode != "RGBA":
        image = image.convert("RGBA")
    if tiled_watermark.mode != "RGBA":
        tiled_watermark = tiled_watermark.convert("RGBA")

    watermarked_image = Image.alpha_composite(image, tiled_watermark)

    return watermarked_image


@cache
def _watermark_image() -> ImageObject:
    return gen_watermark_image("EPAM DIAL")


def stamp_watermark(
    image_bytes: bytes, output_format: Literal["JPEG", "PNG"]
) -> bytes:
    image = Image.open(io.BytesIO(image_bytes))

    watermarked_image = _stamp_watermark_image(image, _watermark_image())

    output_bytes = io.BytesIO()
    watermarked_image.save(output_bytes, format=output_format)

    return output_bytes.getvalue()



from .adapter import interceptor_to_embeddings
from .base import EmbeddingsInterceptor, EmbeddingsNoOpInterceptor

__all__ = [
    "EmbeddingsInterceptor",
    "EmbeddingsNoOpInterceptor",
    "interceptor_to_embeddings",
]



from aidial_interceptors_sdk.examples.app_factory import create_app
from aidial_interceptors_sdk.examples.registry import EXAMPLE_INTERCEPTORS
from aidial_interceptors_sdk.utils._env import get_env
from aidial_interceptors_sdk.utils._http_client import get_http_client

dial_url = get_env("DIAL_URL")


async def client_factory():
    return get_http_client()


app = create_app(
    dial_url=dial_url,
    client_factory=client_factory,
    interceptors=EXAMPLE_INTERCEPTORS,
)



from aidial_sdk.chat_completion import Response
from aidial_sdk.chat_completion.chunks import BaseChunk
from aidial_sdk.pydantic_v1 import PrivateAttr

from aidial_interceptors_sdk.chat_completion.annotated_value import (
    AnnotatedChunk,
)
from aidial_interceptors_sdk.chat_completion.element_path import (
    ChoiceContext,
    ElementPath,
)
from aidial_interceptors_sdk.chat_completion.helpers import (
    traverse_dict_value,
    traverse_list,
)
from aidial_interceptors_sdk.chat_completion.index_mapper import IndexMapper
from aidial_interceptors_sdk.chat_completion.response_message_handler import (
    ResponseMessageHandler,
)
from aidial_interceptors_sdk.utils._dial_sdk import send_chunk_to_response
from aidial_interceptors_sdk.utils.not_given import NotGiven


class ResponseHandler(ResponseMessageHandler):
    """
    Callbacks for handling chat completion responses.

    1. the callbacks may mutate the response in place,
    2. the callbacks are applied in bottom-up order (the order of callbacks in the class definition reflects the order of application),
    3. a callback for a list element return a list of objects, which allows for adding or removing elements. Such callbacks have the type signature: `T -> List[T]`,
    4. a callback for a dictionary value returns an optional dictionary, which allows for adding or removing keys. Such callbacks have the type signature: `(T | NotGiven | None) -> (T | NotGiven | None)`.
    5. `on_response*` methods are expected to be overridden by subclasses.
    6. `traverse_response*` methods are not expected to be overridden by subclasses.

    The callbacks are intended to be used for:
    1. collecting information about the response,
    2. inplace modifying the response.

    Avoid it creating new entities in the response, creation of deeply nested
    entities is hard if possible.

    E.g. adding a new stage in a response which doesn't have custom content is impossible,
    because one would have to create "custom_content" field _before_
    creating the "stages" field. Bottom-up application order of callbacks
    doesn't allow for that.
    """

    response: Response

    # NOTE: `_stage_indices = {}` isn't going to work, since
    # the underscored field `_stage_indices` will be shared across
    # all instances of the class.
    _stage_indices: dict[int, IndexMapper[int]] = PrivateAttr({})

    def _get_stage_index_mapper(self, choice_idx: int) -> IndexMapper[int]:
        if choice_idx not in self._stage_indices:
            self._stage_indices[choice_idx] = IndexMapper()
        return self._stage_indices[choice_idx]

    def reserve_stage_index(self, choice_idx: int) -> int:
        return self._get_stage_index_mapper(choice_idx).reserve()

    def send_chunk(self, chunk: BaseChunk | dict):
        return send_chunk_to_response(self.response, chunk)

    async def on_response_message(
        self, path: ElementPath, message: dict | NotGiven | None
    ) -> dict | NotGiven | None:
        return message

    async def on_response_finish_reason(
        self, path: ElementPath, finish_reason: str | NotGiven | None
    ) -> str | NotGiven | None:
        return finish_reason

    async def on_response_choice(
        self, path: ElementPath, choice: dict
    ) -> list[dict] | dict:
        return choice

    async def on_response_choices(
        self, choices: list[dict] | NotGiven | None
    ) -> list[dict] | NotGiven | None:
        return choices

    # TODO: add path to the signature and to the rest of similar methods
    async def on_response_usage(
        self, usage: dict | NotGiven | None
    ) -> dict | NotGiven | None:
        return usage

    async def on_stream_chunk(self, chunk: dict) -> None:
        self.send_chunk(chunk)

    async def traverse_response_chunk(self, ann_chunk: AnnotatedChunk) -> None:
        r = ann_chunk.chunk

        async def traverse_message(
            path: ElementPath, message: dict | NotGiven | None
        ) -> dict | NotGiven | None:
            if message is not None and not isinstance(message, NotGiven):
                message = await self.traverse_response_message(path, message)
            return await self.on_response_message(path, message)

        async def traverse_choice(
            path: ElementPath, choice: dict
        ) -> list[dict] | dict:
            choice = await traverse_dict_value(
                path, choice, "finish_reason", self.on_response_finish_reason
            )
            choice = await traverse_dict_value(
                path, choice, "delta", traverse_message
            )
            return await self.on_response_choice(path, choice)

        async def traverse_choices(
            path: ElementPath, choices: list[dict] | NotGiven | None
        ) -> list[dict] | NotGiven | None:
            def with_choice_ctx(choice_idx: int) -> ElementPath:
                return path.with_choice_ctx(
                    ChoiceContext(
                        index=choice_idx,
                        stage_index_mapper=self._get_stage_index_mapper(
                            choice_idx
                        ),
                    )
                )

            choices = await traverse_list(
                with_choice_ctx, choices, traverse_choice
            )
            return await self.on_response_choices(choices)

        async def traverse_response_usage(
            path: ElementPath, usage: dict | NotGiven | None
        ) -> dict | NotGiven | None:
            return await self.on_response_usage(usage)

        path = ElementPath(response_ctx=ann_chunk.annotation)
        r = await traverse_dict_value(path, r, "usage", traverse_response_usage)
        r = await traverse_dict_value(path, r, "choices", traverse_choices)

        await self.on_stream_chunk(r)



import inspect
from collections.abc import Awaitable, Callable
from typing import Any, TypeVar

from aidial_sdk.exceptions import InvalidRequestError

T = TypeVar("T")


async def call_with_extra_body(
    func: Callable[..., Awaitable[T]], arg: dict
) -> T:
    if _has_kwargs_argument(func):
        return await func(**arg)

    expected_args = set(inspect.signature(func).parameters.keys())
    actual_args = set(arg.keys())

    extra_args = actual_args - expected_args

    if extra_args and "extra_body" not in expected_args:
        raise InvalidRequestError(
            f"Unrecognized request argument supplied: {extra_args}."
        )

    arg["extra_body"] = arg.get("extra_body") or {}
    arg["model"] = arg.get("model")  # model arg is required in AsyncAzureOpenAI

    for extra_arg in extra_args:
        arg["extra_body"][extra_arg] = arg[extra_arg]
        del arg[extra_arg]

    return await func(**arg)


def _has_kwargs_argument(func: Callable[..., Awaitable[Any]]) -> bool:
    """
    Determines if the given function accepts
    a variable keyword argument (**kwargs).
    """
    signature = inspect.signature(func)
    for param in signature.parameters.values():
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            print(param)
            return True
    return False



import time
from collections import defaultdict

from aidial_sdk.chat_completion import Stage
from aidial_sdk.pydantic_v1 import BaseModel
from typing_extensions import override

from aidial_interceptors_sdk.chat_completion.base import (
    ChatCompletionInterceptor,
)
from aidial_interceptors_sdk.chat_completion.element_path import ElementPath
from aidial_interceptors_sdk.examples.utils.dict import collect_at_path
from aidial_interceptors_sdk.examples.utils.markdown import MarkdownTable
from aidial_interceptors_sdk.utils.not_given import NotGiven


class UsagePerModel(BaseModel):
    index: int
    model: str
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


class StatisticsReporterInterceptor(ChatCompletionInterceptor):
    """
    Reports token usage, finish reason and generation speed in the first stage.
    """

    # Request data
    request_n: int = -1

    # Response data
    request_end_time: float = -1

    response_start_time: float = -1
    response_end_time: float = -1

    prompt_tokens: int = 0
    completion_tokens: int = 0

    usage_per_model: list[UsagePerModel] = []

    # Per-choice response data
    finish_reasons: dict[int, str] = {}
    statistics_stages: dict[int, Stage] = {}
    content_lengths: dict[int, int] = defaultdict(int)

    @override
    async def on_response_usage(
        self, usage: dict | NotGiven | None
    ) -> dict | NotGiven | None:
        if usage:
            self.prompt_tokens += usage["prompt_tokens"]
            self.completion_tokens += usage["completion_tokens"]
        return usage

    @override
    async def on_response_finish_reason(
        self, path: ElementPath, finish_reason: str | NotGiven | None
    ) -> str | NotGiven | None:
        if finish_reason and path.choice_idx is not None:
            self.finish_reasons[path.choice_idx] = finish_reason
        return finish_reason

    @override
    async def on_response_message(
        self, path: ElementPath, message: dict | NotGiven | None
    ) -> dict | NotGiven | None:
        if (
            message
            and path.choice_idx is not None
            and (stage := self.statistics_stages.get(path.choice_idx))
            is not None
            and (content := message.get("content")) is not None
        ):
            stage.append_content(f"`{content or '∅'}`║")
            self.content_lengths[path.choice_idx] += len(content)
        return message

    def _collect_data_points(self, chunk: dict) -> None:
        self.response_end_time = time.time()

        for usage in collect_at_path(chunk, "statistics.usage_per_model.*"):
            self.usage_per_model.append(UsagePerModel.parse_obj(usage))

    def _get_choice_metrics(self, choice_idx: int) -> str:
        # NOTE: each metric but content_length and finish_reason
        # is shared across all choices.
        # So they will be repeated for each choice.

        metrics = MarkdownTable(headers=["Metric", "Value"])

        metrics.add_rows(
            ["Finish reason", self.finish_reasons.get(choice_idx) or "NA"],
            ["Prompt tokens", self.prompt_tokens],
            ["Completion tokens", self.completion_tokens],
        )

        response_duration = self.response_end_time - self.response_start_time
        latency = self.response_start_time - self.request_end_time

        metrics.add_rows(["Latency, sec", f"{latency:.2f}"])
        metrics.add_rows(["Stream duration, sec", f"{response_duration:.2f}"])

        if response_duration > 0:
            metrics.add_rows(
                [
                    "Chars/sec",
                    f"{self.content_lengths[choice_idx] / response_duration:.2f}",
                ],
                [
                    "Tokens/sec",
                    f"{self.completion_tokens / response_duration:.2f}",
                ],
            )

        usages = MarkdownTable(
            title="Per model usage",
            headers=["Idx", "Model", "Prompt tokens", "Completion tokens"],
        )

        for idx, usage in enumerate(self.usage_per_model, start=1):
            usages.add_row(
                [
                    idx,
                    usage.model,
                    usage.prompt_tokens,
                    usage.completion_tokens,
                ]
            )

        return "\n\n" + metrics.to_markdown() + usages.to_markdown_opt()

    @override
    async def on_request(self, request: dict) -> dict:
        self.request_end_time = time.time()
        self.request_n = request.get("n") or 1
        return request

    @override
    async def on_stream_start(self) -> None:
        # NOTE: one could create new stages inside `on_response_choice` callback.
        # However, this will affect the stage indices mapping in case if the
        # first chunk already contains stages. Those will be pinned and new stages
        # will come after them.
        # This is not what we want - we want these stages to *always* come first.
        for choice_idx in range(self.request_n):
            stage = Stage(
                self.response._queue,
                choice_idx,
                self.reserve_stage_index(choice_idx),
                "Statistics",
            )

            stage.open()
            stage.append_content("### Stream\n\n")
            self.statistics_stages[choice_idx] = stage

    @override
    async def on_stream_chunk(self, chunk: dict) -> None:
        if self.response_start_time == -1:
            self.response_start_time = time.time()

        self._collect_data_points(chunk)
        self.send_chunk(chunk)

    @override
    async def on_stream_end(self) -> None:
        for choice_idx in range(self.request_n):
            if stage := self.statistics_stages.get(choice_idx):
                stage.append_content(self._get_choice_metrics(choice_idx))
                stage.close()



import functools
from collections.abc import Awaitable, Callable

import httpx

# connect timeout and total timeout
DEFAULT_TIMEOUT = httpx.Timeout(600, connect=10)

# Borrowed from openai._constants.DEFAULT_CONNECTION_LIMITS
DEFAULT_CONNECTION_LIMITS = httpx.Limits(
    max_connections=1000, max_keepalive_connections=100
)

HTTPClientFactory = Callable[[], Awaitable[httpx.AsyncClient]]


@functools.cache
def get_http_client() -> httpx.AsyncClient:
    return httpx.AsyncClient(
        timeout=DEFAULT_TIMEOUT,
        limits=DEFAULT_CONNECTION_LIMITS,
        follow_redirects=True,
    )



import logging
from collections.abc import AsyncIterator, Callable
from typing import Any, TypeVar

from aidial_sdk.exceptions import HTTPException as DialException

from aidial_interceptors_sdk.chat_completion.annotated_value import (
    AnnotatedChunk,
    AnnotatedException,
    AnnotatedValue,
    Annotation,
)
from aidial_interceptors_sdk.utils._exceptions import to_dial_exception

_log = logging.getLogger(__name__)

_T = TypeVar("_T")
_V = TypeVar("_V")


async def materialize_streaming_errors(
    stream: AsyncIterator[dict],
) -> AsyncIterator[dict | DialException]:
    try:
        async for chunk in stream:
            yield chunk
    except Exception as e:
        _log.exception(
            "caught exception while streaming: "
            f"{type(e).__module__}.{type(e).__name__}"
        )

        yield to_dial_exception(e)


def annotate_stream(
    annotation: Annotation, stream: AsyncIterator[dict | DialException]
) -> AsyncIterator[AnnotatedValue]:
    def _annotate(value: dict | DialException) -> AnnotatedValue:
        if isinstance(value, dict):
            return AnnotatedChunk(chunk=value, annotation=annotation)
        else:
            return AnnotatedException(error=value, annotation=annotation)

    return map_stream(_annotate, stream)


# TODO: add to SDK as a inverse of cleanup_indices
def _add_indices(chunk: Any) -> Any:
    if isinstance(chunk, list):
        ret = []
        for idx, elem in enumerate(chunk, start=1):
            if isinstance(elem, dict) and "index" not in elem:
                elem = {**elem, "index": idx}
            ret.append(_add_indices(elem))
        return ret

    if isinstance(chunk, dict):
        return {key: _add_indices(value) for key, value in chunk.items()}

    return chunk


# TODO: add to SDK as an inverse of merge_chunks
def block_response_to_streaming_chunk(response: dict) -> dict:
    for choice in response["choices"]:
        choice["delta"] = choice["message"]
        del choice["message"]
        _add_indices(choice["delta"])
    return response


async def map_stream(
    func: Callable[[_T], _V | None], iterator: AsyncIterator[_T]
) -> AsyncIterator[_V]:
    async for item in iterator:
        new_item = func(item)
        if new_item is not None:
            yield new_item


async def singleton_stream(item: _T) -> AsyncIterator[_T]:
    yield item



"""
Callbacks to handle messages in the chat completion request.
"""

from abc import ABC

from aidial_sdk.pydantic_v1 import BaseModel

from aidial_interceptors_sdk.chat_completion.element_path import ElementPath
from aidial_interceptors_sdk.chat_completion.helpers import (
    traverse_dict_value,
    traverse_list,
)
from aidial_interceptors_sdk.utils.not_given import NotGiven


class RequestMessageHandler(ABC, BaseModel):
    class Config:
        arbitrary_types_allowed = True

    async def on_request_stage(
        self, path: ElementPath, stage: dict
    ) -> list[dict] | dict:
        return stage

    async def on_request_stages(
        self, path: ElementPath, stages: list[dict] | NotGiven | None
    ) -> list[dict] | NotGiven | None:
        return stages

    async def on_request_attachment(
        self, path: ElementPath, attachment: dict
    ) -> list[dict] | dict:
        """
        Applied to individual attachments: every message and stage attachment
        """

        return attachment

    async def on_request_attachments(
        self, path: ElementPath, attachments: list[dict] | NotGiven | None
    ) -> list[dict] | NotGiven | None:
        """
        Applied to list of attachments: message attachments and stage attachments
        """
        return attachments

    async def on_request_state(
        self, path: ElementPath, state: dict | NotGiven | None
    ) -> dict | NotGiven | None:
        return state

    async def on_request_custom_content(
        self, path: ElementPath, custom_content: dict | NotGiven | None
    ) -> dict | NotGiven | None:
        return custom_content

    async def traverse_request_message(
        self, path: ElementPath, message: dict
    ) -> dict:
        async def apply_on_attachments(
            path: ElementPath, attachments: list[dict] | NotGiven | None
        ) -> list[dict] | NotGiven | None:
            attachments = await traverse_list(
                path.with_attachment_idx,
                attachments,
                self.on_request_attachment,
            )
            return await self.on_request_attachments(path, attachments)

        async def apply_on_stage(
            path: ElementPath, stage: dict
        ) -> list[dict] | dict:
            stage = await traverse_dict_value(
                path, stage, "attachments", apply_on_attachments
            )

            if path.stage_idx is not None and path.choice_ctx is not None:
                mapper = path.choice_ctx.stage_index_mapper
                stage["index"] = mapper(path.stage_idx)

            return await self.on_request_stage(path, stage)

        async def apply_on_stages(
            path: ElementPath, stages: list[dict] | NotGiven | None
        ) -> list[dict] | NotGiven | None:
            stages = await traverse_list(
                path.with_stage_idx, stages, apply_on_stage
            )
            return await self.on_request_stages(path, stages)

        async def apply_on_custom_content(
            path: ElementPath, cc: dict | NotGiven | None
        ) -> dict | NotGiven | None:
            cc = await traverse_dict_value(
                path, cc, "state", self.on_request_state
            )
            cc = await traverse_dict_value(
                path, cc, "attachments", apply_on_attachments
            )
            cc = await traverse_dict_value(path, cc, "stages", apply_on_stages)
            return await self.on_request_custom_content(path, cc)

        return await traverse_dict_value(
            path, message, "custom_content", apply_on_custom_content
        )



import re
from typing import Any

from aidial_sdk.pydantic_v1 import BaseModel


class MarkdownTable(BaseModel):
    title: str | None = None
    headers: list[Any]
    rows: list[list[Any]] = []

    def add_row(self, row: list[Any]) -> None:
        if len(row) != len(self.headers):
            raise ValueError(
                f"Number of headers ({len(self.headers)}) does not match number of cells in a row ({len(row)})"
            )
        self.rows.append(row)

    def add_rows(self, *rows: list[Any]) -> None:
        for row in rows:
            self.add_row(row)

    def is_empty(self) -> bool:
        return not self.rows

    def to_markdown(self) -> str:
        rows = [self.headers, ["---"] * len(self.headers), *self.rows]
        ret = "\n".join(
            f"|{'|'.join(map(escape_table_cell, row))}|" for row in rows
        )
        if self.title:
            ret = f"\n\n### {self.title}\n\n" + ret
        return ret

    def to_markdown_opt(self) -> str:
        if self.is_empty():
            return ""
        return self.to_markdown()


def escape_table_cell(value: Any) -> str:
    return _escape_new_lines(_escape_code_blocks(str(value)))


def _escape_new_lines(text: str) -> str:
    return text.replace("\n", "<br>")


def _escape_code_blocks(text: str) -> str:
    in_code_block = False
    ret = []

    open_regexp = re.compile(r"^```([^\s`]+)?\s*$")
    close_regexp = re.compile(r"^```\s*$")

    for line in text.splitlines(keepends=True):
        if open_regexp.match(line) and not in_code_block:
            in_code_block = True
            ret.append("<pre>")
        elif close_regexp.match(line) and in_code_block:
            in_code_block = False
            ret.append("</pre>")
        else:
            ret.append(line)

    return "".join(ret)



from aidial_sdk.exceptions import InvalidRequestError
from typing_extensions import override

from aidial_interceptors_sdk.embeddings.base import EmbeddingsInterceptor


class BlacklistedWordsInterceptor(EmbeddingsInterceptor):
    BLACKLIST = {"hello", "world"}

    @override
    async def modify_input(self, input: str) -> str:
        for word in self.BLACKLIST:
            if word.lower() in input.lower():
                message = "The embedding input contains a blacklisted word."
                raise InvalidRequestError(
                    message=message,
                    display_message=message,
                )
        return input



"""
All kinds of logic which is a good candidate
to be moved eventually to the SDK itself.
"""

import json
import logging
from typing import TypeVar

import pydantic
from aidial_sdk.chat_completion import Request, Response
from aidial_sdk.chat_completion.chunks import ArbitraryChunk, BaseChunk
from aidial_sdk.exceptions import RequestValidationError
from pydantic import BaseModel

from aidial_interceptors_sdk.utils._string import decapitalize

_log = logging.getLogger(__name__)


def send_chunk_to_response(response: Response, chunk: BaseChunk | dict):
    if isinstance(chunk, dict):
        for choice in chunk.get("choices") or []:
            if (index := choice.get("index")) is not None:
                response._last_choice_index = max(
                    response._last_choice_index, index + 1
                )
        response._queue.put_nowait(ArbitraryChunk(chunk=chunk))
    else:
        response._queue.put_nowait(chunk)


_CONFIGURATION_KEY = "interceptor_configuration"


def cleanup_interceptor_configuration(request_body: dict) -> dict:
    """Leave no traces of the interceptor configuration in the request to the upstream."""

    if (cf := request_body.get("custom_fields")) is not None:
        cf.pop(_CONFIGURATION_KEY, None)
        if not cf:
            request_body.pop("custom_fields")

    return request_body


_T = TypeVar("_T", bound=BaseModel)


def parse_interceptor_configuration(
    request: Request, cls: type[_T] | None
) -> _T | None:
    config: dict | None = None
    if cf := request.custom_fields:
        config = cf.dict().get(_CONFIGURATION_KEY)

    if config is not None:
        _log.debug(f"interceptor configuration: {json.dumps(config)}")

    match (cls, config):
        case (None, None):
            return None
        case (None, _):
            raise RequestValidationError(
                f"The interceptor doesn't have configuration, but it was provided in the chat completion request. Path: 'custom_fields.{_CONFIGURATION_KEY}'"
            )
        case (_, _):
            try:
                conf = {} if config is None else config
                return cls.parse_obj(conf)
            except pydantic.ValidationError as e:
                error = e.errors()[0]
                path = ".".join(map(str, error["loc"]))
                msg = f"Invalid request. Path: 'custom_fields.{_CONFIGURATION_KEY}.{path}', error: {decapitalize(error['msg'])}"

                raise RequestValidationError(msg) from None



from aidial_sdk.exceptions import InternalServerError

from ..anonymizer.base import Anonymizer
from ..anonymizer.replacement import Replacements, create_indexed_replacements
from .client import DlpClient
from .config import DeIdentificationConfig


class GoogleDLPAnonymizer(Anonymizer):
    _client: DlpClient
    _config: DeIdentificationConfig

    def __init__(self, project: str, config: DeIdentificationConfig):
        self._client = DlpClient(project, config)
        self._config = config

    async def collect_replacements(
        self, text: str, *, replacements: Replacements | None = None
    ) -> Replacements:
        anonymized = await self._client.anonymize(text)

        replacements = replacements or Replacements()
        replacements = create_indexed_replacements(
            replacements, self._config.info_types, text, anonymized
        )
        if replacements is None:
            raise InternalServerError(
                "It wasn't possible to parse the anonymized text"
            )
        return replacements



import logging
import os
from collections.abc import Callable
from typing import TypeVar

_log = logging.getLogger(__name__)


def get_env(name: str, err_msg: str | None = None) -> str:
    if (val := os.getenv(name)) is not None:
        return val

    raise Exception(err_msg or f"{name} env variable is not set")


def get_env_list(name: str, default: list[str] | None = None) -> list[str]:
    if (value := os.getenv(name)) is None:
        return default or []
    return value.split(",")


_T = TypeVar("_T")


def get_envs(names: list[str], parser: Callable[[str], _T], default: _T) -> _T:
    for name in names:
        if os.getenv(name) is not None:
            if name != names[-1]:
                _log.warning(
                    f"The environment variable {name!r} is deprecated. "
                    f"Use {names[-1]!r} instead."
                )
            return parser(name)
    return default



import uuid

from aidial_sdk.pydantic_v1 import BaseModel, PrivateAttr

SESSION_ID_KEY = "langfuse_session_id"


class Session(BaseModel):
    session_id: str = ""
    _added_session_id: bool = PrivateAttr(False)

    @classmethod
    def create(cls, messages: list[dict]) -> "Session":
        messages = list(
            filter(
                lambda msg: msg.get("custom_content")
                and msg["custom_content"].get("state")
                and SESSION_ID_KEY in msg["custom_content"]["state"],
                messages,
            )
        )
        if messages:
            session_id = messages[0]["custom_content"]["state"][SESSION_ID_KEY]
        else:
            session_id = str(uuid.uuid4())
        return cls(session_id=session_id)

    def add_session_id_to_message(self, message: dict) -> dict:
        if not self._added_session_id:
            message.setdefault("custom_content", {}).setdefault("state", {})[
                SESSION_ID_KEY
            ] = self.session_id
            self._added_session_id = True
        return message

    def remove_session_id_from_message(self, message: dict) -> dict:
        if (
            message.get("custom_content", {})
            .get("state", {})
            .get(SESSION_ID_KEY)
        ):
            del message["custom_content"]["state"][SESSION_ID_KEY]
        if message.get("custom_content", {}).get("state") == {}:
            del message["custom_content"]
        return message

    def remove_session_id_from_messages(
        self, messages: list[dict]
    ) -> list[dict]:
        return [self.remove_session_id_from_message(msg) for msg in messages]



import json
import logging
import time
import uuid

from aidial_sdk.utils.merge_chunks import merge
from typing_extensions import override

from aidial_interceptors_sdk.chat_completion.base import (
    ChatCompletionInterceptor,
)
from aidial_interceptors_sdk.error import EarlyStreamExit
from aidial_interceptors_sdk.examples.utils.lru_cache import LRUCache

_log = logging.getLogger(__name__)

_MAX_CACHE_SIZE = 1000
_LRU_CACHE = LRUCache[str, dict](maxsize=_MAX_CACHE_SIZE)


def _request_to_key(request: dict) -> str:
    return json.dumps(request, sort_keys=True, separators=(",", ":"))


def _merge_chunks(chunk1: dict, chunk2: dict) -> dict:
    # Avoid merging top-level keys
    for key in ["id", "created", "model", "object", "system_fingerprint"]:
        chunk1[key] = chunk1.get(key) or chunk2.get(key)
        if key in chunk2:
            del chunk2[key]

    return merge(chunk1, chunk2)


class CachingInterceptor(ChatCompletionInterceptor):
    request_key: str = ""
    response_merged: dict = {}

    @override
    async def on_request(self, request: dict) -> dict:
        self.request_key = _request_to_key(request)
        return request

    @override
    async def on_stream_start(self) -> None:
        cached_response = _LRU_CACHE.lookup(self.request_key)
        if cached_response is not None:
            cached_response["id"] = "chatcmpl-" + str(uuid.uuid4())
            cached_response["created"] = int(time.time())
            self.send_chunk(cached_response)
            _log.debug("Cache hit")
            raise EarlyStreamExit("Cache hit")

        _log.debug("Cache miss")

    @override
    async def on_stream_chunk(self, chunk: dict) -> None:
        self.response_merged = _merge_chunks(self.response_merged, chunk)
        self.send_chunk(chunk)

    @override
    async def on_stream_end(self) -> None:
        del self.response_merged["id"]
        del self.response_merged["created"]

        _LRU_CACHE.save(self.request_key, self.response_merged)
        _log.debug("Saved to cache")



from urllib.parse import urljoin

from aidial_sdk.exceptions import InvalidRequestError
from typing_extensions import override

from aidial_interceptors_sdk.chat_completion.base import (
    ChatCompletionInterceptor,
)


class RejectExternalLinksInterceptor(ChatCompletionInterceptor):
    @override
    async def on_request_attachment(self, path, attachment: dict) -> list[dict]:
        url = attachment.get("url")
        if url is None:
            return [attachment]

        dial_url = self.dial_client.dial_url
        abs_url = urljoin(dial_url, url)
        if not abs_url.startswith(dial_url):
            message = f"External links are not allowed: {url!r}"
            raise InvalidRequestError(
                message=message,
                display_message=message,
            )

        return [attachment]



import logging
from functools import cache

from aidial_sdk.pydantic_v1 import BaseModel
from spacy import load as load_model
from spacy.cli.download import download as download_model
from spacy.language import Language

from aidial_interceptors_sdk.utils._env import get_env_list, get_envs

from ..anonymizer.base import Anonymizer
from ..anonymizer.replacement import Replacement
from ..anonymizer.replacements import Replacements

_log = logging.getLogger(__name__)

# Find spaCy models here: https://spacy.io/models/
# NOTE: Pinning the version of en_core_web_sm:
# https://github.com/explosion/spaCy/issues/13690#issuecomment-2487873386
# otherwise, there is a chance of running into 403 error in runtime.
_DEFAULT_MODEL = "en_core_web_sm-3.7.1"

# Find the full list of entities here:
# https://github.com/explosion/spacy-models/blob/e46017f5c8241096c1b30fae080f0e0709c8038c/meta/en_core_web_sm-3.7.0.json#L121-L140
_DEFAULT_LABELS_TO_REDACT = [
    "PERSON",
    "ORG",
    "GPE",  # Geo-political entity
    "PRODUCT",
]


@cache
def _get_pipeline(model: str) -> Language:
    model_name = model.partition("-")[0]  # dropping a version
    try:
        return load_model(model_name)
    except Exception as e:
        _log.warning(
            f"Failed to load spaCy model {model!r}: {str(e)}\nDownloading the model..."
        )
        download_model(model, direct=True)
        return load_model(model_name)


# Preemptively load the default model on the server start-up
# to avoid waiting during the first request.
_get_pipeline(_DEFAULT_MODEL)

_LABELS_TO_REDACT = get_envs(
    [
        "PII_ANONYMIZER_LABELS_TO_REDACT",
        "SPACY_ANONYMIZER_LABELS_TO_REDACT",
    ],
    get_env_list,
    _DEFAULT_LABELS_TO_REDACT,
)


class SpacyAnonymizer(BaseModel, Anonymizer):
    def _is_replacement(self, text: str, start: int, end: int) -> bool:
        return bool(
            Replacement.parse(text[start:end])
            or Replacement.parse(
                text[max(0, start - 1) : min(end + 1, len(text))]
            )
        )

    async def collect_replacements(
        self, text: str, *, replacements: Replacements | None = None
    ) -> Replacements:
        doc = _get_pipeline(_DEFAULT_MODEL)(text)

        replacements = replacements or Replacements()
        for ent in doc.ents:
            if ent.label_ in _LABELS_TO_REDACT and not self._is_replacement(
                doc.text, ent.start_char, ent.end_char
            ):
                replacements.get_replacement(ent.label_, ent.text)

        return replacements



from collections.abc import AsyncIterator, Awaitable, Callable
from typing import (
    TypeVar,
)

import aiostream
from aidial_sdk.chat_completion import Stage
from aidial_sdk.chat_completion.chunks import (
    ContentChunk,
    EndChoiceChunk,
    FinishReason,
    UsageChunk,
)
from aidial_sdk.exceptions import HTTPException as DialException
from typing_extensions import override

from aidial_interceptors_sdk.chat_completion.annotated_value import (
    AnnotatedValue,
    Annotation,
)
from aidial_interceptors_sdk.chat_completion.base import (
    ChatCompletionInterceptor,
    RequestDict,
)
from aidial_interceptors_sdk.chat_completion.element_path import ElementPath
from aidial_interceptors_sdk.chat_completion.index_mapper import IndexMapper
from aidial_interceptors_sdk.utils.not_given import NotGiven
from aidial_interceptors_sdk.utils.streaming import annotate_stream


class ReplicatorInterceptor(ChatCompletionInterceptor):
    """
    Call the upstream model N time and combine the results into a single response.
    """

    # Interceptor parameter
    n: int

    # Response data
    content_stages: dict[int, Stage] = {}
    stage_index_mapper: IndexMapper[tuple[int, int]] = IndexMapper()

    total_usage: UsageChunk = UsageChunk(0, 0, None)
    finish_reasons: dict[int, str] = {}
    role_sent: bool = False
    chunk_template: dict = {}

    def _accumulate_usage(self, usage: dict) -> None:
        self.total_usage.prompt_tokens += usage.get("prompt_tokens") or 0
        self.total_usage.completion_tokens += (
            usage.get("completion_tokens") or 0
        )

    def _get_content_stage(self, path: ElementPath) -> Stage:
        response_idx = path.response_ctx
        assert isinstance(response_idx, int)

        if response_idx not in self.content_stages:
            stage = Stage(
                self.response._queue,
                0,
                self.stage_index_mapper((response_idx, -1)),
                f"{response_idx + 1} | CONTENT",
            )
            stage.open()
            self.content_stages[response_idx] = stage

        return self.content_stages[response_idx]

    @override
    async def call_upstreams(
        self,
        request: RequestDict,
        call_upstream: Callable[
            [Annotation, RequestDict],
            Awaitable[AsyncIterator[dict | DialException]],
        ],
    ) -> AsyncIterator[AnnotatedValue]:
        request["n"] = 1

        streams = [
            annotate_stream(idx, await call_upstream(idx, request))
            for idx in range(self.n)
        ]
        # TODO: create tasks
        return _join_iterators(streams)

    @override
    async def on_response_stage(
        self, path: ElementPath, stage: dict
    ) -> list[dict] | dict:
        assert path.stage_idx is not None
        assert isinstance(path.response_ctx, int)

        if name := stage.get("name"):
            stage["name"] = f"{path.response_ctx + 1} | {name}"

        stage["index"] = self.stage_index_mapper(
            (path.response_ctx, path.stage_idx)
        )

        return [stage]

    @override
    async def on_response_message(
        self, path: ElementPath, message: dict | NotGiven | None
    ) -> dict | NotGiven | None:
        if message:
            # Avoid sending role multiple times
            if message.get("role"):
                if self.role_sent:
                    del message["role"]
                else:
                    self.role_sent = True

            # Rerouting message content to a dedicated content stage
            if content := message.get("content"):
                self._get_content_stage(path).append_content(content)
                message["content"] = ""

            if (cc := message.get("custom_content")) and (
                attachments := cc.get("attachments")
            ):
                for attachment in attachments:
                    self._get_content_stage(path).add_attachment(**attachment)
                del cc["attachments"]

        return message

    @override
    async def on_response_finish_reason(
        self, path: ElementPath, finish_reason: str | NotGiven | None
    ) -> str | NotGiven | None:
        assert isinstance(path.response_ctx, int)

        if finish_reason:
            self.finish_reasons[path.response_ctx + 1] = finish_reason

        return None

    @override
    async def on_response_usage(
        self, usage: dict | NotGiven | None
    ) -> dict | NotGiven | None:
        # TODO: combine statistics.usage_per_model
        if usage:
            self._accumulate_usage(usage)

        return None

    @override
    async def on_stream_chunk(self, chunk: dict) -> None:
        if not self.chunk_template:
            self.chunk_template = {
                k: v
                for k, v in chunk.items()
                if k in ["id", "created", "system_fingerprint"]
            }

        self.send_chunk(self.chunk_template | chunk)

    # TODO: support streaming errors

    @override
    async def on_stream_end(self) -> None:
        content = "\n\n".join(
            [
                "Total usage:",
                f" * Prompt tokens: {self.total_usage.prompt_tokens}",
                f" * Completion tokens: {self.total_usage.completion_tokens}",
                "Finish reasons:",
                f" * {str(self.finish_reasons)}",
            ]
        )
        self.send_chunk(ContentChunk(content, 0))

        self.send_chunk(self.total_usage)

        for stage in self.content_stages.values():
            stage.close()

        finish_reason = (
            FinishReason.STOP
            if not self.finish_reasons
            else FinishReason(self.finish_reasons[1])
        )
        self.send_chunk(EndChoiceChunk(finish_reason, 0))


_T = TypeVar("_T")


async def _join_iterators(iters: list[AsyncIterator[_T]]) -> AsyncIterator[_T]:
    async with aiostream.stream.merge(*iters).stream() as combine:
        async for item in combine:
            yield item



from collections.abc import AsyncIterator, Awaitable, Callable
from typing import TypeVar

from aidial_sdk.exceptions import HTTPException as DialException
from pydantic import BaseModel

from aidial_interceptors_sdk.chat_completion.annotated_value import (
    AnnotatedException,
    AnnotatedValue,
    Annotation,
)
from aidial_interceptors_sdk.chat_completion.request_handler import (
    RequestHandler,
)
from aidial_interceptors_sdk.chat_completion.response_handler import (
    ResponseHandler,
)
from aidial_interceptors_sdk.dial_client import DialClient
from aidial_interceptors_sdk.utils.streaming import annotate_stream

RequestDict = dict

_T = TypeVar("_T", bound=BaseModel)


class ChatCompletionInterceptor(RequestHandler, ResponseHandler):
    dial_client: DialClient
    configuration: BaseModel | None

    @classmethod
    async def configuration_schema(cls) -> type[BaseModel] | None:
        return None

    def get_configuration(self, cls: type[_T]) -> _T:
        if self.configuration is None:
            raise ValueError("Can't find interceptor configuration")

        if not isinstance(self.configuration, cls):
            raise TypeError(
                f"The interceptor configuration is of type {type(self.configuration).__name__!r}, but expected to be of type {cls.__name__!r}"
            )

        return self.configuration

    async def call_upstreams(
        self,
        request: RequestDict,
        call_upstream: Callable[
            [Annotation, RequestDict],
            Awaitable[AsyncIterator[dict | DialException]],
        ],
    ) -> AsyncIterator[AnnotatedValue]:
        annotation = None
        return annotate_stream(
            annotation, await call_upstream(annotation, request)
        )

    async def on_stream_start(self) -> None:
        # TODO: it's probably worth to put all the chunks
        # generated by this method into a separate list.
        # And then merge them all with the first incoming chunk.
        # Otherwise, we may end up with choice being open *before*
        # its "assistant" role is reported.
        pass

    async def on_stream_error(self, error: AnnotatedException) -> None:
        raise error.error

    async def on_stream_end(self) -> None:
        # TODO: it's probably worth to withhold the last chunk generated by
        # on_stream_chunk and merge it with all the chunks reported by on_stream_end.
        pass


class ChatCompletionNoOpInterceptor(ChatCompletionInterceptor):
    pass



import base64

import numpy as np


def vector_to_base64(vector: list[float]) -> str:
    array = np.array(vector, dtype="float32")
    byte_data = array.tobytes()
    base64_encoded = base64.b64encode(byte_data).decode("utf-8")
    return base64_encoded


def base64_to_vector(data: str) -> list[float]:
    return np.frombuffer(base64.b64decode(data), dtype="float32").tolist()



from pydantic import BaseModel

from aidial_interceptors_sdk.utils._env import get_env_list

_DEFAULT_INFO_TYPES = get_env_list(
    "GOOGLE_DLP_ANONYMIZER_INFO_TYPES_TO_DE_IDENTIFY",
    ["PHONE_NUMBER", "FIRST_NAME", "LAST_NAME"],
)


class DeIdentificationConfig(BaseModel):
    class Config:
        extra = "allow"

    info_types: list[str] = _DEFAULT_INFO_TYPES


class GoogleDLPAnonymizerConfig(BaseModel):
    class Config:
        extra = "allow"

    deidentification_config: DeIdentificationConfig = DeIdentificationConfig()



from ..anonymizer.base import Anonymizer
from ..anonymizer.impl import AnonymizerInterceptor
from .anonymizer import SpacyAnonymizer


class SpacyAnonymizerInterceptor(AnonymizerInterceptor):
    def get_anonymizer(self) -> Anonymizer:
        return SpacyAnonymizer()



from aidial_sdk import DIALApp
from aidial_sdk.telemetry.types import TelemetryConfig

from aidial_interceptors_sdk.chat_completion import (
    interceptor_to_chat_completion,
)
from aidial_interceptors_sdk.embeddings.adapter import interceptor_to_embeddings
from aidial_interceptors_sdk.examples.registry import Interceptors
from aidial_interceptors_sdk.examples.utils.log_config import configure_loggers
from aidial_interceptors_sdk.utils._http_client import HTTPClientFactory


def create_app(
    *,
    dial_url: str,
    client_factory: HTTPClientFactory,
    interceptors: Interceptors,
) -> DIALApp:
    app = DIALApp(
        description="Examples of DIAL interceptors",
        dial_url=dial_url,
        telemetry_config=TelemetryConfig(),
        add_healthcheck=True,
        propagate_auth_headers=True,
        allow_extra_request_fields=True,
    )

    configure_loggers()

    for id, cls in interceptors.embeddings.items():
        app.add_embeddings(
            id, interceptor_to_embeddings(cls, dial_url, client_factory)
        )

    for id, cls in interceptors.chat_completions.items():
        app.add_chat_completion(
            id, interceptor_to_chat_completion(cls, dial_url, client_factory)
        )

    return app



from typing_extensions import override

from aidial_interceptors_sdk.chat_completion.base import (
    ChatCompletionInterceptor,
)


class PirateInterceptor(ChatCompletionInterceptor):
    @override
    async def on_request_messages(self, messages: list[dict]) -> list[dict]:
        instruction = "Reply as a pirate."
        if len(messages) > 0 and messages[0]["role"] == "system":
            messages[0]["content"] = messages[0]["content"] + f"\n{instruction}"
            return messages
        else:
            return [{"role": "system", "content": instruction}, *messages]



from collections.abc import Mapping

from aidial_sdk.exceptions import InvalidRequestError
from aidial_sdk.pydantic_v1 import BaseModel
from openai import AsyncAzureOpenAI

from aidial_interceptors_sdk.utils._http_client import HTTPClientFactory
from aidial_interceptors_sdk.utils.storage import FileStorage

_UPSTREAMS_HEADER = "X-UPSTREAMS"


class DialClient(BaseModel):
    dial_url: str
    client: AsyncAzureOpenAI
    storage: FileStorage

    class Config:
        arbitrary_types_allowed = True

    @classmethod
    async def create(
        cls,
        *,
        dial_url: str,
        api_key: str | None,
        authorization: str | None,
        api_version: str | None,
        headers: Mapping[str, str],
        client_factory: HTTPClientFactory,
    ) -> "DialClient":
        if not api_key:
            raise InvalidRequestError("The 'api-key' request header is missing")

        extra_headers: dict[str, str] = {}
        if authorization is not None:
            extra_headers["Authorization"] = authorization

        azure_deployment = "interceptor"
        if upstreams_header := headers.get(_UPSTREAMS_HEADER):
            azure_deployment, *upstreams = upstreams_header.split(
                ",", maxsplit=1
            )
            if upstreams:
                extra_headers[_UPSTREAMS_HEADER] = "".join(upstreams)

        http_client = await client_factory()

        client = AsyncAzureOpenAI(
            azure_endpoint=dial_url,
            azure_deployment=azure_deployment,
            # NOTE: DIAL SDK takes care of propagating api-key header
            api_key="-",
            # NOTE: api-version query parameter is not required in the chat completions DIAL API.
            # However, it is required in Azure OpenAI API, that's why the openai library fails when it's missing:
            # https://github.com/openai/openai-python/blob/9850c169c4126fd04dc6796e4685f1b9e4924aa4/src/openai/lib/azure.py#L174-L177
            #
            # A workaround for it could be to patch the AsyncAzureOpenAI class in order to disable the check.
            # This would be hard to maintain though.
            #
            # However, since DIAL OpenAI adapter treats a missing api-version in the same way as an empty string,
            # we could actually default the api-version to an empty string here too.
            # DIAL OpenAI adapter is the only place where api-version has any effect,
            # so the query param modification is safe.
            # https://github.com/epam/ai-dial-adapter-openai/blob/b462d1c26ce8f9d569b9c085a849206aad91becf/aidial_adapter_openai/app.py#L93
            api_version=api_version or "",
            max_retries=0,
            http_client=http_client,
            # NOTE: if Authorization header was provided in the request,
            # then propagate it to the upstream.
            # Whether interceptor gets the header or not, is determined by
            # `forwardAuthToken` option set for the interceptor in the DIAL Core config.
            default_headers=extra_headers,
        )

        storage = FileStorage(
            dial_url=dial_url,
            api_key=api_key,
            http_client=http_client,
        )

        return cls(
            dial_url=dial_url,
            client=client,
            storage=storage,
        )



import functools
import json
import logging
from collections.abc import Awaitable, Callable
from typing import TypeVar

_log = logging.getLogger(__name__)


_A = TypeVar("_A")
_B = TypeVar("_B")


def debug_logging(
    title: str,
) -> Callable[
    [Callable[[_A], Awaitable[_B]]],
    Callable[[_A], Awaitable[_B]],
]:
    def decorator(
        func: Callable[[_A], Awaitable[_B]],
    ) -> Callable[[_A], Awaitable[_B]]:
        if not _log.isEnabledFor(logging.DEBUG):
            return func

        @functools.wraps(func)
        async def wrapper(a: _A) -> _B:
            _log.debug(f"{title} old: {json.dumps(a)}")
            b = await func(a)
            _log.debug(f"{title} new: {json.dumps(b)}")
            return b

        return wrapper

    return decorator



import logging
from abc import ABC

from aidial_sdk.pydantic_v1 import BaseModel

from aidial_interceptors_sdk.dial_client import DialClient

_log = logging.getLogger(__name__)


class EmbeddingsInterceptor(ABC, BaseModel):
    class Config:
        arbitrary_types_allowed = True

    dial_client: DialClient

    async def modify_input(self, input: str) -> str:
        return input

    async def modify_embedding(
        self, embedding: str | list[float]
    ) -> str | list[float]:
        return embedding

    async def modify_request(self, request: dict) -> dict:
        if "input" in request:
            input = request["input"]
            if isinstance(input, str):
                request["input"] = await self.modify_input(input)
            elif isinstance(input, list):
                if all(isinstance(item, str) for item in input):
                    request["input"] = [
                        await self.modify_input(item) for item in input
                    ]
            else:
                _log.warning("Tokenized input isn't yet supported")

        return request

    async def modify_response(self, response: dict) -> dict:
        for item in response.get("data") or []:
            item["embedding"] = await self.modify_embedding(item["embedding"])
        return response


class EmbeddingsNoOpInterceptor(EmbeddingsInterceptor):
    pass



from aidial_sdk.embeddings import Embeddings
from aidial_sdk.embeddings.request import Request
from aidial_sdk.embeddings.response import Response
from openai.types import CreateEmbeddingResponse

from aidial_interceptors_sdk.dial_client import DialClient
from aidial_interceptors_sdk.embeddings.base import EmbeddingsInterceptor
from aidial_interceptors_sdk.utils._debug import debug_logging
from aidial_interceptors_sdk.utils._exceptions import dial_exception_decorator
from aidial_interceptors_sdk.utils._http_client import HTTPClientFactory
from aidial_interceptors_sdk.utils._reflection import call_with_extra_body


def interceptor_to_embeddings(
    cls: type[EmbeddingsInterceptor],
    dial_url: str,
    client_factory: HTTPClientFactory,
) -> Embeddings:
    class Impl(Embeddings):
        @dial_exception_decorator
        async def embeddings(self, request: Request) -> Response:
            dial_client = await DialClient.create(
                dial_url=dial_url,
                api_key=request.api_key,
                api_version=request.api_version,
                authorization=request.jwt,
                headers=request.headers,
                client_factory=client_factory,
            )

            interceptor = cls(
                dial_client=dial_client,
                **request.original_request.path_params,
            )

            body = await request.original_request.json()
            body = await debug_logging("request")(interceptor.modify_request)(
                body
            )

            response: CreateEmbeddingResponse = await call_with_extra_body(
                dial_client.client.embeddings.create, body
            )

            response_dict = await debug_logging("response")(
                interceptor.modify_response
            )(response.to_dict())

            return Response.parse_obj(response_dict)

    return Impl()



from datetime import datetime

import httpx
from aidial_sdk.utils.merge_chunks import merge
from typing_extensions import override

from aidial_interceptors_sdk.chat_completion.base import (
    ChatCompletionInterceptor,
)
from aidial_interceptors_sdk.chat_completion.element_path import ElementPath
from aidial_interceptors_sdk.examples.chat_completion.langfuse.langfuse_client import (
    LangfuseClient,
)
from aidial_interceptors_sdk.examples.chat_completion.langfuse.session import (
    Session,
)
from aidial_interceptors_sdk.utils.not_given import NotGiven


class LangfuseInterceptor(ChatCompletionInterceptor):
    """
    Save data to Langfuse
    """

    session: Session = Session()
    request_deployment_id: str = ""
    request_model: str = ""
    merged_response_message = {}
    start_time: datetime | None
    x_conversation_id: str = ""

    @override
    async def on_response_message(
        self, path: ElementPath, message: dict | NotGiven | None
    ) -> dict | NotGiven | None:
        if message:
            message = self.session.add_session_id_to_message(message)
            self.merged_response_message = merge(
                self.merged_response_message, message
            )
        return message

    @override
    async def on_request(self, request: dict) -> dict:
        self.request_deployment_id = self.request.deployment_id or ""
        self.request_model = self.request.model or ""
        self.x_conversation_id = self.request.headers["x-conversation-id"]
        self.session = Session.create(messages=request["messages"])
        request["messages"] = self.session.remove_session_id_from_messages(
            messages=request["messages"]
        )
        return request

    @override
    async def on_stream_start(self) -> None:
        self.start_time = datetime.now()

    @override
    async def on_stream_end(self) -> None:
        model_info = await self._get_model_info()
        user_email = await self._get_user_email()
        is_model = bool(model_info)
        tags = []
        if self.request_model:
            tags.append(self.request_model)
        if self.request_deployment_id:
            tags.append(self.request_deployment_id)
        LangfuseClient(
            session_id=self.session.session_id,
            tags=tags,
            request_messages=self.request.messages,
            response_message=self.merged_response_message,
            model_name=self.request_model,
            deployment_id=self.request_deployment_id,
            start_time=self.start_time or datetime.now(),
            end_time=datetime.now(),
            user_id=user_email,
            metadata={
                "model": self.request_model,
                "deployment_id": self.request_deployment_id,
                "x_conversation_id": self.x_conversation_id,
                "is_model": is_model,
                "model_info": model_info,
            },
            is_model=is_model,
        ).transmit()

    async def _get_user_email(self) -> str:
        response = await self._get_dial_client().get("/v1/user/info")
        response.raise_for_status()
        return response.json().get("userClaims", {}).get("email", [""])[0]

    async def _get_model_info(self) -> dict | None:
        response = await self._get_dial_client().get("/openai/models")
        response.raise_for_status()
        model_info = next(
            (
                item
                for item in response.json().get("data", [])
                if item["id"] == self.request_model
            ),
            None,
        )
        return model_info

    def _get_dial_client(self) -> httpx.AsyncClient:
        return httpx.AsyncClient(
            base_url=self.dial_client.dial_url,
            headers={"Api-Key": self.dial_client.storage.api_key},
        )



from aidial_sdk.exceptions import InvalidRequestError
from typing_extensions import override

from aidial_interceptors_sdk.chat_completion.base import (
    ChatCompletionInterceptor,
)
from aidial_interceptors_sdk.utils.not_given import NotGiven


class BlacklistedWordsInterceptor(ChatCompletionInterceptor):
    BLACKLIST = {"hello", "world"}

    total_content: str = ""

    def _validate_content_in(self, entity: str, content: str):
        for word in self.BLACKLIST:
            if word.lower() in content.lower():
                error_message = (
                    f"The chat completion {entity} contains a blacklisted word."
                )
                raise InvalidRequestError(
                    message=error_message,
                    display_message=error_message,
                )

    @override
    async def on_request_message(self, path, message: dict) -> list[dict]:
        content = message.get("content") or ""
        self._validate_content_in("request", content)
        return [message]

    @override
    async def on_response_message(
        self, path, message: dict | NotGiven | None
    ) -> dict | NotGiven | None:
        if isinstance(message, dict):
            content = message.get("content") or ""
            self.total_content += content
            self._validate_content_in("response", self.total_content)
        return message



from collections.abc import Callable
from typing import Any, TypeVar

Path = list[str | int]
PathLike = Path | str


def _parse_path(path: str) -> Path:
    ret: Path = []
    for segment in path.split("."):
        if not segment:
            raise ValueError("Invalid empty path segment.")
        if segment[0] == "!":
            ret.append(int(segment[1:]))
        else:
            ret.append(segment)

    return ret


def _to_path(path: Path | str) -> Path:
    if isinstance(path, str):
        return _parse_path(path)
    else:
        return path


def _update_at_path(
    obj: Any,
    path: Path,
    acc_path: Path,
    fn: Callable[[Path, Any], Any],
    **kwargs: Any,
) -> Any:
    """
    Immutable modification of a nested value in a dictionary or list.

    If a dict key is missing, then it's created, unless drill_down is False.
    If list/tuple index is missing, then an error is raised.
    Catch-all "*" index is supported for lists.
    """

    drill_down = kwargs.get("drill_down", False)
    strict = kwargs.get("strict", True)

    if path == []:
        return fn(acc_path, obj)

    key = path[0]
    rest = path[1:]

    if isinstance(obj, dict):
        value = obj.get(key, None)
        if value is None and not drill_down:
            return obj
        else:
            return {
                **obj,
                key: _update_at_path(
                    value, rest, acc_path + [key], fn, **kwargs
                ),
            }

    if isinstance(obj, list):
        if key == "*":
            indices = set(range(len(obj)))
        else:
            if isinstance(key, int):
                index = key
            elif strict:
                raise ValueError(f"Invalid list index: {key!r}")
            else:
                return obj

            if index < 0 or index >= len(obj) and strict:
                raise ValueError(
                    f"Invalid index ({index}) for a list of length {len(obj)}"
                )
            indices = {index}

        return [
            (
                _update_at_path(elem, rest, acc_path + [i], fn, **kwargs)
                if i in indices
                else elem
            )
            for i, elem in enumerate(obj)
        ]

    if isinstance(obj, tuple):
        if isinstance(key, int):
            index = key
        elif strict:
            raise ValueError(f"Invalid tuple index: {key!r}")
        else:
            return obj

        if index < 0 or index >= len(obj) and strict:
            raise ValueError(
                f"Invalid index ({index}) for a tuple of length {len(obj)}"
            )

        return tuple(
            (
                _update_at_path(elem, rest, acc_path + [i], fn, **kwargs)
                if i == index
                else elem
            )
            for i, elem in enumerate(obj)
        )

    if obj is None:
        if isinstance(key, int):
            if strict:
                raise ValueError("Cannot index into None")
            else:
                return obj
        else:
            if drill_down:
                return {
                    key: _update_at_path(
                        obj, rest, acc_path + [key], fn, **kwargs
                    )
                }
            else:
                return obj

    if strict:
        raise ValueError(
            f"Cannot get a value at path {path} in a scalar value of type {type(obj).__name__}"
        )
    else:
        return obj


def update_at_path(
    obj: Any, path: PathLike, fn: Callable[[Path, Any], Any], **kwargs
) -> Any:
    return _update_at_path(obj, _to_path(path), [], fn, **kwargs)


def set_at_path(obj: dict, path: PathLike, value: Any, **kwargs) -> dict:
    return update_at_path(obj, path, lambda path, _value: value, **kwargs)


T = TypeVar("T")


def get_at_path(obj: Any, path: PathLike) -> Any:
    values = collect_at_path(obj, path)
    return None if values == [] else values[-1]


def collect_at_path(obj: Any, path: PathLike) -> list[Any]:
    ret = []

    def fn(path, value):
        ret.append(value)
        return value

    update_at_path(obj, path, fn)
    return ret
